# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gtk, Gio, WebKit2, Gdk

from gettext import gettext as _
from urllib.parse import urlparse
from time import time

from eolie.define import App, LoadingType, LoadingState
from eolie.utils import get_ftp_cmd, emit_signal
from eolie.logger import Logger


class WebViewNavigation:
    """
        Implement WebView navigation (uri, title, readable, ...)
        Should be inherited by a WebView
    """

    __MIMES = ["text/html", "text/xml", "application/xhtml+xml",
               "x-scheme-handler/http", "x-scheme-handler/https",
               "multipart/related", "application/x-mimearchive",
               "application/x-extension-html"]

    def __init__(self):
        """
            Init navigation
        """
        self.__loaded_uri = None
        self.__insecure_content_detected = False
        self.connect("decide-policy", self.__on_decide_policy)
        self.connect("insecure-content-detected",
                     self.__on_insecure_content_detected)
        self.connect("run-as-modal", self.__on_run_as_modal)
        self.connect("permission_request", self.__on_permission_request)

    def load_uri(self, uri):
        """
            Load uri
            @param uri as str
        """
        parsed = urlparse(uri)
        # If not an URI, start a search
        is_uri = parsed.scheme in ["about", "http",
                                   "https", "file", "populars"]
        if not is_uri and\
                not uri.startswith("/") and\
                App().search.is_search(uri):
            uri = App().search.get_search_uri(uri)
        parsed = urlparse(uri)
        if uri == "about:blank":
            WebKit2.WebView.load_plain_text(self, "")
        # We are not a ftp browser, fall back to env
        elif parsed.scheme == "ftp":
            argv = [get_ftp_cmd(), uri, None]
            GLib.spawn_sync(None, argv, None,
                            GLib.SpawnFlags.SEARCH_PATH, None)
        else:
            if uri.startswith("/"):
                uri = "file://" + uri
            elif parsed.scheme == "javascript":
                # To bypass popup blocker
                self._last_click_time = time()
                uri = GLib.uri_unescape_string(uri, None)
                self.run_javascript(uri.replace("javascript:", ""), None, None)
            elif parsed.scheme not in ["http", "https", "file",
                                       "populars", "accept"]:
                uri = "http://" + uri
            # Reset bad tls certificate
            if parsed.scheme != "accept":
                self.reset_bad_tls()
                self.__insecure_content_detected = False
            self.stop_loading()
            self.set_uri(uri)
            self.__loaded_uri = self.uri
            GLib.idle_add(WebKit2.WebView.load_uri, self, uri)

    @property
    def loaded_uri(self):
        """
            Get loaded uri
            @return str/None
        """
        return self.__loaded_uri

#######################
# PROTECTED           #
#######################
    def _set_user_agent(self, uri):
        """
            Set user agent for uri
            @param uri as str
        """
        user_agent = App().websettings.get("user_agent", uri)
        settings = self.get_settings()
        if user_agent:
            settings.set_user_agent(user_agent)
        else:
            settings.set_user_agent_with_application_details("Eolie",
                                                             None)

    def _on_load_changed(self, webview, event):
        """
            Update internals
            @param webview as WebView
            @param event as WebKit2.LoadEvent
        """
        parsed = urlparse(webview.uri)
        if event == WebKit2.LoadEvent.STARTED:
            self._loading_state = LoadingState.LOADING
        elif event == WebKit2.LoadEvent.COMMITTED:
            if parsed.scheme in ["http", "https"]:
                emit_signal(self, "title-changed", webview.uri)
                self.update_zoom_level()
                self.update_sound_policy()
        elif event == WebKit2.LoadEvent.FINISHED:
            if self._loading_state not in [LoadingState.STOPPED,
                                           LoadingState.ERROR]:
                self._loading_state = LoadingState.NONE
            App().history.set_page_state(self.uri)
            self.__update_bookmark_metadata(self.uri)
            self.update_spell_checking(self.uri)
            if App().show_tls:
                try:
                    from OpenSSL import crypto
                    from datetime import datetime
                    (valid, tls, errors) = webview.get_tls_info()
                    if tls is not None:
                        Logger.info("***************************************"
                                    "***************************************")
                        cert_pem = tls.get_property("certificate-pem")
                        cert = crypto.load_certificate(crypto.FILETYPE_PEM,
                                                       cert_pem)
                        subject = cert.get_subject()
                        Logger.info("CN: %s", subject.CN)
                        start_bytes = cert.get_notBefore()
                        end_bytes = cert.get_notAfter()
                        start = datetime.strptime(start_bytes.decode("utf-8"),
                                                  "%Y%m%d%H%M%SZ")
                        end = datetime.strptime(end_bytes.decode("utf-8"),
                                                "%Y%m%d%H%M%SZ")
                        Logger.info("Valid from %s to %s", (start, end))
                        Logger.info("Serial number: %s",
                                    cert.get_serial_number())
                        Logger.info(cert_pem)
                        Logger.info("***************************************"
                                    "***************************************")
                except Exception as e:
                    Logger.info("Please install OpenSSL python support: %s", e)

#######################
# PRIVATE             #
#######################
    def __update_bookmark_metadata(self, uri):
        """
            Update bookmark access time/popularity
            @param uri as str
        """
        if App().bookmarks.get_id(uri) is not None:
            App().bookmarks.set_access_time(uri, round(time(), 2))
            App().bookmarks.set_more_popular(uri)

    def __on_run_as_modal(self, webview):
        Logger.info("WebView::__on_run_as_modal(): TODO")

    def __on_insecure_content_detected(self, webview, event):
        """
            @param webview as WebView
            @param event as WebKit2.InsecureContentEvent
        """
        self.__insecure_content_detected = True

    def __on_permission_request(self, webview, request):
        """
            Handle Webkit permissions
            @param webview as WebKit2.WebView
            @param request as WebKit2.PermissionRequest
        """
        def on_allow_notifications(parsed):
            values = list(App().settings.get_value("notification-domains"))
            values.append("%s;%s" % (parsed.scheme, parsed.netloc))
            App().settings.set_value("notification-domains",
                                     GLib.Variant("as", values))
            request.allow()

        if isinstance(request, WebKit2.GeolocationPermissionRequest):
            if self.is_ephemeral:
                request.deny()
            else:
                uri = webview.uri
                self.window.toolbar.title.show_geolocation(uri, request)
        elif isinstance(request, WebKit2.NotificationPermissionRequest):
            parsed = urlparse(webview.uri)
            from eolie.container_notification import ContainerNotification
            notification = ContainerNotification(
                _("Allow notifications for %s") % parsed.netloc,
                [_("Allow")],
                [lambda: on_allow_notifications(parsed)])
            self.window.container.overlay.add_overlay(notification)
            notification.show()
            notification.set_reveal_child(True)
        return True

    def __on_decide_policy(self, webview, decision, decision_type):
        """
            Navigation policy
            @param webview as WebKit2.WebView
            @param decision as WebKit2.NavigationPolicyDecision
            @param decision_type as WebKit2.PolicyDecisionType
            @return bool
        """
        # Always accept response
        if decision_type == WebKit2.PolicyDecisionType.RESPONSE:
            response = decision.get_response()
            mime_type = response.props.mime_type
            uri = response.get_uri()
            parsed = urlparse(uri)
            if mime_type in self.__MIMES:
                decision.use()
                return False
            elif parsed.scheme == "file":
                f = Gio.File.new_for_uri(uri)
                info = f.query_info("standard::type",
                                    Gio.FileQueryInfoFlags.NOFOLLOW_SYMLINKS,
                                    None)
                if info.get_file_type() == Gio.FileType.REGULAR:
                    try:
                        Gtk.show_uri_on_window(self.window,
                                               uri,
                                               Gtk.get_current_event_time())
                    except Exception as e:
                        Logger.error("""WebViewNavigation::
                                        __on_decide_policy(): %s""", e)
                    decision.ignore()
                    return True
                else:
                    decision.use()
                    return False
            elif self.can_show_mime_type(mime_type):
                decision.use()
                return False
            else:
                decision.download()
                return True

        navigation_action = decision.get_navigation_action()
        navigation_uri = navigation_action.get_request().get_uri()
        mouse_button = navigation_action.get_mouse_button()
        parsed_navigation = urlparse(navigation_uri)
        self.clear_text_entry()
        if parsed_navigation.scheme not in ["http", "https", "file", "about",
                                            "populars", "accept"]:
            try:
                Gtk.show_uri_on_window(self.window,
                                       navigation_uri,
                                       Gtk.get_current_event_time())
            except Exception as e:
                Logger.error("WebViewNavigation::__on_decide_policy(): %s", e)
            decision.ignore()
        elif mouse_button == 0:
            # Prevent opening empty pages
            if navigation_uri == "about:blank":
                self.reset_last_click_event()
                decision.use()
                return False
            elif decision_type == WebKit2.PolicyDecisionType.NEW_WINDOW_ACTION:
                decision.use()
                return False
            else:
                decision.use()
                return False
        elif mouse_button == 1:
            modifiers = navigation_action.get_modifiers()
            if decision_type == WebKit2.PolicyDecisionType.NEW_WINDOW_ACTION:
                decision.use()
                return False
            elif modifiers == Gdk.ModifierType.CONTROL_MASK:
                self.new_page(navigation_uri, LoadingType.BACKGROUND)
                decision.ignore()
                return True
            elif modifiers == Gdk.ModifierType.SHIFT_MASK:
                self.new_page(navigation_uri, LoadingType.POPOVER)
                decision.ignore()
                return True
            else:
                self.__loaded_uri = navigation_uri.rstrip("/")
                emit_signal(self, "title-changed", navigation_uri)
                decision.use()
                return False
        else:
            self.new_page(navigation_uri, LoadingType.BACKGROUND)
            decision.ignore()
            return True



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib


class MessagePopover(Gtk.Popover):
    """
        Show  message
        @warning: will block current execution
    """

    def __init__(self, message, window, callback=None, *args):
        """
            Init popover
            @param message as str
            @param window as window
            @param callback as function
        """
        Gtk.Popover.__init__(self)
        self.set_modal(False)
        window.register(self)
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/PopoverJavaScript.ui")
        widget = builder.get_object("widget")
        label = builder.get_object("label")
        image = builder.get_object("image")
        image.set_from_icon_name("dialog-warning-symbolic",
                                 Gtk.IconSize.DIALOG)
        label.set_text(message)
        self.add(widget)
        self.__loop = GLib.MainLoop.new(None, False)
        self.connect("closed", self.__on_closed)
        if callback is not None:
            self.__callback = callback
            self.__args = args
            ok_button = builder.get_object("ok_button")
            cancel_button = builder.get_object("cancel_button")
            ok_button.show()
            cancel_button.show()
            builder.connect_signals(self)

    def popup(self):
        """
            Popup widget and run loop
        """
        Gtk.Popover.popup(self)
        self.__loop.run()

#######################
# PROTECTED           #
#######################
    def _on_ok_button_clicked(self, button):
        """
            Pass ok to js
            @param button as Gtk.Button
        """
        self.__callback(*self.__args)
        self.hide()

    def _on_cancel_button_clicked(self, button):
        """
            Pass ok to js
            @param button as Gtk.Button
        """
        self.hide()

#######################
# PRIVATE             #
#######################
    def __on_closed(self, popover):
        """
            Quit main loop
            @param popover as Gtk.Popover
        """
        self.__loop.quit()
        self.__loop.unref()



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib

from gettext import gettext as _
from datetime import datetime

from eolie.define import MARGIN


class TLSPopover(Gtk.Popover):
    """
        Show TLS information
    """

    def __init__(self, window):
        """
            Init popover
            @param window as Window
        """
        Gtk.Popover.__init__(self)
        self.get_style_context().add_class("dark")
        self.set_modal(False)
        window.register(self)
        grid = Gtk.Grid.new()
        grid.show()
        grid.set_row_spacing(MARGIN)
        grid.set_orientation(Gtk.Orientation.VERTICAL)
        title_label = Gtk.Label.new()
        title_label.show()
        title_label.get_style_context().add_class("bold")
        title_label.get_style_context().add_class("dim-label")
        cert_label = Gtk.Label.new()
        cert_label.show()
        cert_label.set_property("halign", Gtk.Align.START)
        grid.add(title_label)
        grid.add(cert_label)
        grid.get_style_context().add_class("padding")
        webview = window.container.webview
        try:
            from OpenSSL import crypto
            (valid, tls, error) = webview.get_tls_info()
            if valid:
                text = ""
                title_label.set_text(_("The connexion is secure"))
                cert_pem = tls.get_property("certificate-pem")
                cert = crypto.load_certificate(crypto.FILETYPE_PEM, cert_pem)
                subject = cert.get_subject()
                if subject.CN is not None:
                    text += _("<b>Website identity:</b> %s\n\n") %\
                        GLib.markup_escape_text(subject.CN)
                end_bytes = cert.get_notAfter()
                end = datetime.strptime(end_bytes.decode("utf-8"),
                                        "%Y%m%d%H%M%SZ")
                text += _("<b>Expires on:</b> %s\n\n") % end
                cert_pem = tls.get_issuer().get_property("certificate-pem")
                cert = crypto.load_certificate(crypto.FILETYPE_PEM, cert_pem)
                subject = cert.get_subject()
                if subject.O is not None:
                    text += _("<b>Verified by:</b> %s\n\n") %\
                        GLib.markup_escape_text(subject.O)
                cert_label.set_markup(text)
            else:
                title_label.set_text(_("The connexion is unsecure"))
        except Exception as e:
            title_label.set_text(str(e))
        self.add(grid)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

from eolie.pages_manager_row import PagesManagerRow
from eolie.define import ArtSize, MARGIN


class PagesManagerList(Gtk.ScrolledWindow):
    """
        WebViews in a ListBox
    """

    def __init__(self, window):
        """
            Init Box
            @param window as Window
        """
        Gtk.ScrolledWindow.__init__(self)
        self.__window = window
        self.__selected_row = None
        self.__hovered_row = None
        self.__listbox = Gtk.ListBox.new()
        self.__listbox.show()
        self.__listbox.set_sort_func(self.__sort_func)
        self.__listbox.set_selection_mode(Gtk.SelectionMode.NONE)
        self.__listbox.set_activate_on_single_click(True)
        self.__listbox.connect("row-activated", self.__on_row_activated)
        self.__listbox.get_style_context().add_class("transparent")
        self.__viewport = Gtk.Viewport.new()
        self.add(self.__listbox)
        self.get_style_context().add_class("padding")
        self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)

    def populate(self, webviews):
        """
            Populate list with webviews
            @param webviews as [WebView]
        """
        current_webview = self.__window.container.webview
        for webview in webviews:
            row = PagesManagerRow(webview, self.__window)
            row.show()
            row.connect("destroy", self.__on_row_destroy)
            self.__listbox.add(row)
            self.__calculate_height()
            if webview == current_webview:
                self.__selected_row = row
                row.set_state_flags(Gtk.StateFlags.VISITED, False)

#######################
# PRIVATE             #
#######################
    def __calculate_height(self):
        """
            Update popover height request
        """
        height = 0
        for row in self.__listbox.get_children():
            height += ArtSize.START_HEIGHT + 8 + MARGIN * 2
        size = self.__window.get_size()
        # 8 for child padding/border, 10 for scrolled padding + MARGIN
        # TODO: Use style context to get this
        self.set_size_request(ArtSize.START_WIDTH + 8 + 10 + MARGIN * 2,
                              min(size[1] - 100, height + 10))

    def __sort_func(self, row1, row2):
        """
            Sort pages
            @param row1 as PageChildRow
            @param row2 as PageChildRow
        """
        # Always show current first
        if self.__selected_row in [row1, row2]:
            return self.__selected_row == row2
        else:
            return row2.webview.atime > row1.webview.atime

    def __on_row_destroy(self, row):
        """
            Destroy list if empty
            @parma row as PageRow
        """
        if self.__listbox.get_children():
            self.__calculate_height()
        else:
            popover = self.get_ancestor(Gtk.Popover)
            if popover is not None:
                popover.destroy()

    def __on_row_activated(self, listbox, row):
        """
            Switch to webview
            @param listbox as Gtk.ListBox
            @param row as PageRow
        """
        self.__window.container.set_visible_webview(row.webview)
        self.get_ancestor(Gtk.Popover).hide()



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk, Gio, GLib, WebKit2

from gettext import gettext as _

from eolie.define import App
from eolie.logger import Logger


class FindWidget(Gtk.SearchBar):
    """
        Show a find in webpage widget
    """

    def __init__(self, window):
        """
            Init widget
            @param window as Window
        """
        Gtk.SearchBar.__init__(self)
        self.__find_controller = None
        self.__action = None
        self.__window = window
        self.__count = 0
        self.__current = 0
        self.__search_entry = Gtk.SearchEntry.new()
        self.__search_entry.set_size_request(300, -1)
        self.__search_entry.connect("search-changed", self.__on_search_changed)
        self.__search_entry.connect("key-press-event", self.__on_key_press)
        self.__search_entry.connect("map", self.__on_map)
        self.__search_entry.connect("unmap", self.__on_unmap)
        self.__search_entry.show()
        backward_button = Gtk.Button.new_from_icon_name("go-up-symbolic",
                                                        Gtk.IconSize.BUTTON)
        backward_button.set_tooltip_text(_("Search previous occurrence"))
        backward_button.connect("clicked",
                                lambda x:
                                self.__on_shortcut_action(
                                    None,
                                    GLib.Variant("s", "prev")))
        backward_button.show()
        forward_button = Gtk.Button.new_from_icon_name("go-down-symbolic",
                                                       Gtk.IconSize.BUTTON)
        forward_button.set_tooltip_text(_("Search next occurrence"))
        forward_button.connect("clicked",
                               lambda x:
                               self.__on_shortcut_action(
                                   None,
                                   GLib.Variant("s", "next")))
        forward_button.show()

        self.__label = Gtk.Label.new()
        self.__label.set_property("halign", Gtk.Align.START)
        self.__label.show()

        linked = Gtk.Grid()
        linked.add(self.__search_entry)
        linked.add(backward_button)
        linked.add(forward_button)
        linked.show()
        linked.get_style_context().add_class('linked')
        grid = Gtk.Grid()
        grid.set_column_spacing(5)
        grid.add(linked)
        grid.add(self.__label)
        grid.show()
        self.connect_entry(self.__search_entry)
        self.set_show_close_button(True)
        self.add(grid)

    def search(self):
        """
            Search for current clipboard
        """
        webview = self.__window.container.webview
        webview.run_javascript_from_gresource(
                "/org/gnome/Eolie/javascript/Selection.js", None,
                self.__on_get_selection)

#######################
# PRIVATE             #
#######################
    def __on_counted_matches(self, find_controller, count):
        """
            Update count label
            @param find_controller as WebKit2.FindController
            @param count as str
        """
        self.__count = count
        self.__label.set_text("%s matches" % self.__count)

    def __on_key_press(self, widget, event):
        """
            If Esc, hide widget, why GTK doesn't do that?
            Otherwise, we get an ugly frame
            @param widget as Gtk.SearchEntry
            @param event as Gdk.Event
        """
        if event.keyval == Gdk.KEY_Escape:
            self.__search_entry.set_text("")
            self.set_search_mode(False)
        elif event.keyval == Gdk.KEY_Return:
            self.__find_controller.search_next()
            self.__current += 1

    def __on_map(self, entry):
        """
            Set shortcuts
            @param entry as Gtk.Entry
        """
        self.__label.set_width_chars(len(_("1000 of 1000 matches")))
        self.__action = Gio.SimpleAction.new("find_shortcut",
                                             GLib.VariantType.new('s'))
        self.__action.connect("activate", self.__on_shortcut_action)
        App().add_action(self.__action)
        App().set_accels_for_action("app.find_shortcut::next", ["F3"])
        App().set_accels_for_action("app.find_shortcut::prev", ["<Shift>F3"])
        self.__find_controller =\
            self.__window.container.webview.get_find_controller()
        self.__find_controller.connect("counted-matches",
                                       self.__on_counted_matches)

    def __on_unmap(self, entry):
        """
            Unset shortcuts
            @param entry as Gtk.Entry
        """
        self.__label.set_width_chars(-1)
        App().remove_action("find_shortcut")
        self.__find_controller.search_finish()
        self.__find_controller.disconnect_by_func(self.__on_counted_matches)

    def __on_shortcut_action(self, action, param):
        """
            Global shortcuts handler
            @param action as Gio.SimpleAction/None
            @param param as GLib.Variant
        """
        if self.__count == 0:
            return
        string = param.get_string()
        if string == "next":
            self.__current += 1
            if self.__current > self.__count:
                self.__current = 1
            self.__find_controller.search_next()
        elif string == "prev":
            self.__current -= 1
            if self.__current < 1:
                self.__current = self.__count
            self.__find_controller.search_previous()
        self.__label.set_text("%s of %s matches" % (self.__current,
                                                    self.__count))

    def __on_search_changed(self, entry):
        """
            Update highlight
            @param entry as Gtk.Entry
        """
        text = entry.get_text()
        self.__label.set_text("")
        if len(text) < 2:
            return
        self.__current = 0
        # FIXME Can't understand what is max count :/
        self.__find_controller.count_matches(
            text,
            WebKit2.FindOptions.CASE_INSENSITIVE,
            100)
        self.__find_controller.search(
            text,
            WebKit2.FindOptions.CASE_INSENSITIVE,
            100)

    def __on_get_selection(self, source, result):
        """
            Start search with selection
            @param source as GObject.Object
            @param result as Gio.AsyncResult
        """
        selection = None
        try:
            data = source.run_javascript_from_gresource_finish(result)
            selection = data.get_js_value().to_string()
        except Exception as e:
            Logger.warning("FindWidget::__on_get_selection(): %s", e)
        if selection is None:
            selection = ""
        self.__search_entry.set_text(selection)
        self.__search_entry.grab_focus()



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib
from gi.repository.Gio import FILE_ATTRIBUTE_STANDARD_NAME, \
                              FILE_ATTRIBUTE_TIME_MODIFIED

import json
from time import time

from eolie.content_blocker import ContentBlocker
from eolie.define import ADBLOCK_URIS, App
from eolie.logger import Logger


SCAN_QUERY_INFO = "{},{}".format(FILE_ATTRIBUTE_STANDARD_NAME,
                                 FILE_ATTRIBUTE_TIME_MODIFIED)


class AdContentBlocker(ContentBlocker):
    """
        A WebKit Content Blocker for ads
    """

    def __init__(self):
        """
            Init adblock helper
        """
        try:
            ContentBlocker.__init__(self, "block-ads")
            f = Gio.File.new_for_path(
                    "%s/block-ads.json" % self._JSON_PATH)
            if f.query_exists():
                info = f.query_info(SCAN_QUERY_INFO,
                                    Gio.FileQueryInfoFlags.NONE,
                                    None)
                mtime = int(info.get_attribute_as_string("time::modified"))
            else:
                mtime = 0
            if App().settings.get_value("block-ads"):
                GLib.timeout_add_seconds(7200, self.__download_task, True)
                if time() - mtime > 7200:
                    GLib.timeout_add_seconds(20, self.__download_task, False)
        except Exception as e:
            Logger.error("AdContentBlocker::__init__(): %s", e)

#######################
# PRIVATE             #
#######################
    def __download_task(self, loop):
        """
            Update database from the web, for timeout_add()
            @param loop as bool
        """
        if not Gio.NetworkMonitor.get_default().get_network_metered():
            self.__download_uris(list(ADBLOCK_URIS))
        return loop

    def __download_uris(self, uris, rules=[]):
        """
            Update database from the web
            @param uris as [str]
            @param data as []
        """
        if not Gio.NetworkMonitor.get_default().get_network_available():
            return
        if uris:
            uri = uris.pop(0)
            self._task_helper.load_uri_content(uri,
                                               self._cancellable,
                                               self.__on_load_uri_content,
                                               uris,
                                               rules)

    def __on_load_uri_content(self, uri, status, content, uris, rules):
        """
            Save loaded values
            @param uri as str
            @param status as bool
            @param content as bytes
            @param uris as [str]
            @param rules as []
        """
        try:
            Logger.debug("AdContentBlocker::__on_load_uri_content(): %s", uri)
            if status:
                rules += json.loads(content.decode("utf-8"))
            if uris:
                self.__download_uris(uris, rules)
            elif rules:
                # Save to sources
                f = Gio.File.new_for_path(
                    "%s/block-ads.json" % self._JSON_PATH)
                content = json.dumps(rules).encode("utf-8")
                f.replace_contents(content,
                                   None,
                                   False,
                                   Gio.FileCreateFlags.REPLACE_DESTINATION,
                                   None)
                self._task_helper.run(self._save_rules, rules)
        except Exception as e:
            Logger.error("AdContentBlocker::__on_load_uri_content(): %s", e)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk, WebKit2

from time import time
from uuid import uuid4
from urllib.parse import urlparse

from eolie.helper_passwords import PasswordsHelper
from eolie.define import App
from eolie.utils import get_baseuri
from eolie.logger import Logger


class WebViewCredentials:
    """
        Implement WebView credentials
    """

    def __init__(self):
        """
            Init credentials
        """
        self.__uuid = ""
        self.__in_secrets_user_form_name = ""
        self.__user_form_name = ""
        self.__user_form_value = ""
        self.__pass_form_name = ""
        self.__uri = ""
        self.__form_uri = ""
        self.__ctime = 0
        self.__helper = PasswordsHelper()

    def add_credentials(self, message):
        """
            Add credential to webview
            @param message as str
        """
        try:
            split = message.split("\n")
            if split[1] != "undefined":
                self.__user_form_name = split[1]
                self.__user_form_value = split[2]
                self.__pass_form_name = split[3]
                self.__pass_form_value = split[4]
                self.__form_uri = get_baseuri(split[5])
                self.__hostname = get_baseuri(self.uri)
                self.__ctime = int(time())
                self.__helper.get(self.__hostname,
                                  self.__user_form_name,
                                  self.__pass_form_name,
                                  self.__on_get_password)
        except Exception as e:
            Logger.error("WebViewCredentials::add_credentials(): %s", e)

    def save_credentials(self):
        """
            Save credentials to secrets
        """
        try:
            # Store a new password if non exists or if login is different
            if not self.__uuid or\
                    self.__in_secrets_login != self.__user_form_value:
                self.__uuid = str(uuid4())
                self.__helper.store(self.__user_form_name,
                                    self.__user_form_value,
                                    self.__pass_form_name,
                                    self.__pass_form_value,
                                    self.__hostname,
                                    self.__form_uri,
                                    self.__uuid,
                                    None)
            else:
                self.__helper.clear(self.__uuid,
                                    self.__helper.store,
                                    self.__user_form_name,
                                    self.__user_form_value,
                                    self.__pass_form_name,
                                    self.__pass_form_value,
                                    self.__hostname,
                                    self.__form_uri,
                                    self.__uuid,
                                    None)
            if App().sync_worker is not None:
                App().sync_worker.push_password(self.__user_form_name,
                                                self.__user_form_value,
                                                self.__pass_form_name,
                                                self.__pass_form_value,
                                                self.__hostname,
                                                self.__form_uri,
                                                self.__uuid)
        except Exception as e:
            Logger.error("WebViewCredentials::save_credentials(): %s", e)
        self.__in_secrets_login = ""

    def show_form_menu(self, message):
        """
            Show form menu for message
            @param message as str
        """
        try:
            from eolie.menu_form import FormMenu
            split = message.split("\n")
            user_form_name = split[1]
            menu = FormMenu(self, self.window)
            self.__helper.get(get_baseuri(self.uri), user_form_name, None,
                              self.__on_menu_get_password,
                              menu,
                              user_form_name)
        except Exception as e:
            Logger.error("WebViewCredentials::show_form_menu(): %s", e)

    @property
    def credentials_uri(self):
        """
            Get credentials URI
            @return str
        """
        return self.__hostname

    @property
    def ctime(self):
        """
            Get credentials creation time
            @return int
        """
        return self.__ctime

#######################
# PROTECTED           #
#######################
    def _on_load_changed(self, webview, event):
        """
            Run JS helper
            @param webview as WebView
            @param event as WebKit2.LoadEvent
        """
        if event == WebKit2.LoadEvent.FINISHED:
            self.run_javascript_from_gresource(
                "/org/gnome/Eolie/javascript/Submit.js", None, None)

#######################
# PRIVATE             #
#######################
    def __on_get_password(self, attributes, password, uri, index, count):
        """
            Update uuid
            @param attributes as {}
            @param password as str
            @param uri as str
            @param index as int
            @param count as int
        """
        try:
            if attributes is None:
                return
            if attributes["login"] != self.__user_form_value and\
                    password != self.__pass_form_value:
                self.__in_secrets_login = attributes["login"]
                self.__uuid = attributes["uuid"]
        except Exception as e:
            Logger.error("WebViewCredentials::__on_get_password(): %s", e)

    def __on_menu_get_password(self, attributes, password,
                               uri, index, count, menu, user_form_name):
        """
            Append to menu
            @param attributes as {}
            @param password as str
            @param uri as str
            @param index as int
            @param count as int
            @param menu as FormMenu
            @param user_form_name as str
        """
        try:
            parsed = urlparse(uri)
            if attributes is not None and (
                    count > 0 or parsed.scheme == "http"):
                menu.add_attributes(attributes, uri)
                if index == 0:
                    popover = Gtk.Popover.new_from_model(self, menu)
                    popover.set_modal(False)
                    self.window.register(popover)
                    rect = Gdk.Rectangle()
                    rect.x = self._last_click_event_x
                    rect.y = self._last_click_event_y - 10
                    rect.width = rect.height = 1
                    popover.set_pointing_to(rect)
                    popover.popup()
        except Exception as e:
            Logger.error("WebViewCredentials::__on_get_password(): %s", e)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gio, Gtk, WebKit2

from urllib.parse import urlparse
from gettext import gettext as _

from eolie.logger import Logger
from eolie.define import App, COOKIES_PATH, EOLIE_DATA_PATH, StartPage
from eolie.helper_task import TaskHelper


class Context:
    """
        Handle context signals
    """

    def __init__(self, context):
        """
            Init context
            @param context as WebKit2.WebContext
        """
        self.__context = context
        self.__task_helper = TaskHelper()
        if not context.is_ephemeral():
            context.set_cache_model(WebKit2.CacheModel.WEB_BROWSER)
            context.set_favicon_database_directory(App().favicons_path)
            cookie_manager = context.get_cookie_manager()
            cookie_manager.set_accept_policy(
                App().settings.get_enum("cookie-storage"))
            path = COOKIES_PATH % (EOLIE_DATA_PATH, "default")
            cookie_manager.set_persistent_storage(
                path,
                WebKit2.CookiePersistentStorage.SQLITE)
        context.set_process_model(
            WebKit2.ProcessModel.MULTIPLE_SECONDARY_PROCESSES)
        context.set_spell_checking_enabled(
            App().settings.get_value("enable-spell-check"))
        context.register_uri_scheme("populars", self.__on_populars_scheme)
        context.register_uri_scheme("internal", self.__on_internal_scheme)
        context.register_uri_scheme("accept", self.__on_accept_scheme)
        context.get_security_manager().register_uri_scheme_as_local("populars")
        context.set_sandbox_enabled(True)
        values = App().settings.get_value("notification-domains")
        try:
            origins = []
            for value in values:
                (scheme, netloc) = value.split(";")
                origins.append(WebKit2.SecurityOrigin(scheme, netloc, 0))
            context.initialize_notification_permissions(origins, [])
        except Exception as e:
            Logger.error("Context::__init__(): %s", e)
        # We allow DownloadPopover to connect before default context
        context.connect_after("download-started", self.__on_download_started)

#######################
# PRIVATE             #
#######################
    def __on_populars_scheme(self, request):
        """
            Show populars web pages
            @param request as WebKit2.URISchemeRequest
        """
        uri = request.get_uri()
        parsed = urlparse(uri)
        items = []
        start_page = App().settings.get_enum("start-page")
        wanted = App().settings.get_value("max-popular-items").get_int32()
        if start_page == StartPage.POPULARITY_BOOKMARKS:
            reset_function = "reset_bookmark"
            for (item_id, uri, title) in App().bookmarks.get_populars(wanted):
                items.append((title, uri, "", 1))
        else:
            reset_function = "reset_history"
            for (item_id, uri,
                 netloc, title, count) in App().history.get_populars_by_netloc(
                    parsed.netloc,
                    wanted):
                items.append((title, uri, netloc, count))
        start = Gio.File.new_for_uri("resource:///org/gnome/Eolie/start.html")
        end = Gio.File.new_for_uri("resource:///org/gnome/Eolie/end.html")
        (status, start_content, tag) = start.load_contents(None)
        (status, end_content, tag) = end.load_contents(None)
        # Update start
        html_start = start_content.decode("utf-8")
        html_start = html_start.replace("@TITLE@", _("Popular pages"))
        if App().settings.get_value("night-mode"):
            html_start = html_start.replace("@BACKGROUND_COLOR@",
                                            "#353535")
        else:
            fake = Gtk.Entry.new()
            style_context = fake.get_style_context()
            (found, color) = style_context.lookup_color(
                "theme_selected_bg_color")
            if found:
                color.alpha = 0.2
                html_start = html_start.replace("@BACKGROUND_COLOR@",
                                                color.to_string())
            else:
                html_start = html_start.replace("@BACKGROUND_COLOR@",
                                                "rgba(74,144,217,0.2)")
        idx = 0
        for (title, uri, netloc, count) in items:
            element_id = "element_%s" % idx
            idx += 1
            if count == 1:  # No navigaiftion for one page
                netloc = uri
            if App().settings.get_value("night-mode"):
                path = App().art.get_path(uri, "start_dark")
            else:
                path = App().art.get_path(uri, "start_light")
            if path is None or\
                    not GLib.file_test(path, GLib.FileTest.IS_REGULAR):
                continue
            favicon_path = App().art.get_favicon_path(netloc)
            if favicon_path is not None:
                favicon_uri = "file://%s" % favicon_path
            else:
                favicon_uri = "internal://web-browser-symbolic"
            html_start += '<a class="child color" id="%s"\
                           title="%s" href="%s">\
                           <img src="file://%s"></img>\
                           <footer class="caption color">%s\
                           <img onclick="%s(event, %s, %s)"\
                                class="close_button color">\
                           <img class="favicon" src="%s">\
                           </img></img></footer></a>' % (
                element_id, title, netloc, path,
                title, reset_function,
                "'%s'" % netloc,
                "'%s'" % element_id, favicon_uri)
        html = html_start.encode("utf-8") + end_content
        stream = Gio.MemoryInputStream.new_from_data(html)
        request.finish(stream, -1, "text/html")

    def __on_internal_scheme(self, request):
        """
            Load an internal resource
            @param request as WebKit2.URISchemeRequest
        """
        # We use internal:// because resource:// is already used by WebKit2
        icon_name = request.get_uri().replace("internal://", "")
        icon_info = Gtk.IconTheme.get_default().lookup_icon(
            icon_name, 22,
            Gtk.IconLookupFlags.FORCE_SVG)
        if icon_info is None:
            return
        filename = icon_info.get_filename()
        if filename.endswith(".png"):
            mime = "image/png"
        else:
            mime = "image/svg+xml"
        f = Gio.File.new_for_path(filename)
        request.finish(f.read(), -1, mime)

    def __on_accept_scheme(self, request):
        """
            Accept certificate for uri
            @param request as WebKit2.URISchemeRequest
        """
        view = request.get_web_view()
        if view.bad_tls is None:
            return
        request_uri = request.get_uri()
        parsed = urlparse(request_uri)
        uri = request_uri.replace("accept://", "https://")
        if not App().websettings.get("accept_tls", uri):
            App().websettings.set("accept_tls", uri, True)
        self.__context.allow_tls_certificate_for_host(
            view.bad_tls,
            # Remove port
            parsed.netloc.split(":")[0])
        view.load_uri(uri)

    def __on_download_started(self, context, download):
        """
            A new download started, handle signals
            @param context as WebKit2.WebContext
            @param download as WebKit2.Download
        """
        App().download_manager.add(download)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, WebKit2

from urllib.parse import urlparse

from eolie.define import App


class WebViewContainer:
    """
        WebView management for container
    """

    def __init__(self):
        """
            Init container
        """
        self.__current_webview = None
        self.__signal_ids = []
        self.__bfl_signal_id = None

    def set_visible_webview(self, webview):
        """
            Manage webview signals
            @param webview as WebView
        """
        self.dismiss_webview(self.__current_webview)
        self.__current_webview = webview
        self.__signal_ids.append(
            webview.connect("load-changed", self.__on_load_changed))
        self.__signal_ids.append(
            webview.connect("title-changed", self.__on_title_changed))
        self.__signal_ids.append(
            webview.connect("uri-changed", self.__on_uri_changed))
        self.__signal_ids.append(
            webview.connect("readability-status",
                            self.__on_readability_status))
        self.__signal_ids.append(
            webview.connect("notify::estimated-load-progress",
                            self.__on_estimated_load_progress))
        self.__signal_ids.append(
            webview.connect("enter-fullscreen", self.__on_enter_fullscreen))
        self.__signal_ids.append(
            webview.connect("leave-fullscreen", self.__on_leave_fullscreen))
        self.__signal_ids.append(
            webview.connect("insecure-content-detected",
                            self.__on_insecure_content_detected))
        self.__signal_ids.append(
            webview.connect("mouse-target-changed",
                            self.__on_mouse_target_changed))
        self.__signal_ids.append(
            webview.connect("destroy", self.__on_destroy))
        self.__bfl_signal_id = webview.get_back_forward_list().connect(
                "changed",
                self.__on_back_forward_list_changed)
        accept_tls = App().websettings.get("accept_tls", webview.uri)
        self._window.toolbar.end.show_tls_button(accept_tls)
        self._window.toolbar.actions.set_actions(self.__current_webview)
        self._window.toolbar.title.entry.set_uri(webview.uri)
        self._window.toolbar.title.entry.icons.show_geolocation(False)
        self._window.toolbar.title.entry.icons.show_readable_button(False)
        self._window.toolbar.title.entry.icons.set_loading(False)
        self._window.toolbar.title.entry.progress.hide()
        if webview.get_uri() is None and\
                webview.uri is not None and\
                webview.related is None:
            webview.load_uri(webview.uri)
        if webview.uri == "populars:":
            self._window.toolbar.title.entry.set_default_placeholder()
        else:
            self._window.toolbar.title.entry.set_title(webview.title)

    def dismiss_webview(self, webview):
        """
            Dismiss webview from handlers
            @param webview as WebView
        """
        if self.__current_webview is None or webview != self.__current_webview:
            return
        for signal_id in self.__signal_ids:
            webview.disconnect(signal_id)
        if self.__bfl_signal_id is not None:
            webview.get_back_forward_list().disconnect(self.__bfl_signal_id)
        self.__signal_ids = []
        self.__bfl_signal_id = None
        self.__current_webview = None

#######################
# PRIVATE             #
#######################
    def __on_destroy(self, webview):
        """
            No more signals handlers
            @param webview as WebView
        """
        self.__signal_ids = []
        self.__bfl_signal_id = None
        self.__current_webview = None

    def __on_title_changed(self, webview, title):
        """
            Update title
            @param webview as WebView
            @param title as str
        """
        self._window.toolbar.title.entry.set_title(title)

    def __on_uri_changed(self, webview, uri):
        """
            Update title bar
            @param webview as WebView
            @param uri as str
        """
        if self.reading:
            self.toggle_reading()
        accept_tls = App().websettings.get("accept_tls", uri)
        self._window.toolbar.end.show_tls_button(accept_tls)
        self._window.toolbar.title.entry.set_uri(uri)

    def __on_estimated_load_progress(self, webview, value):
        """
            Update progress bar
            @param webview as WebView
            @param value GparamFloat
        """
        value = webview.get_estimated_load_progress()
        self._window.toolbar.title.entry.progress.set_fraction(value)

    def __on_back_forward_list_changed(self, bf_list, added, removed):
        """
            Update actions
            @param bf_list as WebKit2.BackForwardList
            @param added as WebKit2.BackForwardListItem
            @param removed as WebKit2.BackForwardListItem
            @param webview as WebView
        """
        if self.__current_webview is not None:
            self._window.toolbar.actions.set_actions(self.__current_webview)

    def __on_readability_status(self, webview, status):
        """
            Show/hide toolbar indicator
            @param webview as WebView
            @param status as bool
        """
        self._window.toolbar.title.entry.icons.show_readable_button(status)

    def __on_load_changed(self, webview, event):
        """
            Update UI based on current event
            @param webview as WebView
            @param event as WebKit2.LoadEvent
        """
        parsed = urlparse(webview.uri)
        wanted_scheme = parsed.scheme in ["http", "https", "file"]
        if event == WebKit2.LoadEvent.STARTED:
            self._window.toolbar.title.entry.icons.set_loading(True)
            self._window.toolbar.title.entry.icons.show_geolocation(False)
            self._window.toolbar.title.entry.icons.show_readable_button(False)
        elif event == WebKit2.LoadEvent.COMMITTED:
            if not wanted_scheme:
                GLib.idle_add(self._window.toolbar.title.start_search)
        elif event == WebKit2.LoadEvent.FINISHED:
            self._window.toolbar.title.entry.icons.set_loading(False)
            self._window.toolbar.title.entry.progress.set_fraction(1.0)
            if wanted_scheme:
                GLib.idle_add(webview.grab_focus)
            webview.check_readability()

    def __on_enter_fullscreen(self, webview):
        """
            Go fullscreen
            @param webview as WebView
        """
        self._window.fullscreen(False)

    def __on_leave_fullscreen(self, webview):
        """
            Leave fullscreen
            @param webview as WebView
        """
        self._window.unfullscreen(False)

    def __on_insecure_content_detected(self, webview, event):
        """
            @param webview as WebView
            @param event as WebKit2.InsecureContentEvent
        """
        self._window.toolbar.title.entry.set_insecure_content()

    def __on_mouse_target_changed(self, webview, hit, modifiers):
        """
            Show uri label
            @param webview as WebView
            @param hit as WebKit2.HitTestResult
            @param modifiers as Gdk.ModifierType
        """
        if hit.context_is_link():
            self._uri_label.set_text(hit.get_link_uri())
            self._uri_label.show()
        else:
            self._uri_label.hide()



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib

App = Gio.Application.get_default

PROXY_BUS = 'org.gnome.Eolie.Proxy.Page%s'
PROXY_PATH = '/org/gnome/EolieProxy'
PROXY_INTERFACE = 'org.gnome.Eolie.Proxy'

# Setup common paths
EOLIE_DATA_PATH = GLib.get_user_data_dir() + "/eolie"
EOLIE_CACHE_PATH = GLib.get_user_cache_dir() + "/eolie"

COOKIES_PATH = "%s/cookies_%s.db"

ADBLOCK_URIS = [
 "https://easylist-downloads.adblockplus.org/easylist_content_blocker.json",
 "https://better.fyi/blockerList.json"
]

PHISHING_URI = "https://raw.githubusercontent.com/bellegarde-c/" +\
    "webkit-blocker-phishing/master/phishing.json"

MARGIN = 15
MARGIN_SMALL = 5


class TimeSpan:
    HOUR = "0"
    DAY = "1"
    WEEK = "2"
    FOUR_WEEK = "3"
    YEAR = "4"
    FOREVER = "5"
    NEVER = "6"
    CUSTOM = "7"


TimeSpanValues = {
    TimeSpan.HOUR: GLib.TIME_SPAN_HOUR,
    TimeSpan.DAY: GLib.TIME_SPAN_DAY,
    TimeSpan.WEEK: GLib.TIME_SPAN_DAY * 7,
    TimeSpan.FOUR_WEEK: GLib.TIME_SPAN_DAY * 7 * 4,
    TimeSpan.YEAR: GLib.TIME_SPAN_DAY * 365,
    TimeSpan.FOREVER: 0
}


class ArtSize:
    FAVICON_MIN = 16
    FAVICON = 22
    PREVIEW_WIDTH = 192
    PREVIEW_HEIGHT = 60
    START_WIDTH = 300
    START_HEIGHT = 200


class Indicator:
    NONE = 0
    GEOLOCATION = 1


class LoadingType:
    FOREGROUND = 0
    BACKGROUND = 1
    OFFLOAD = 2
    POPOVER = 3


class Type:
    NONE = -1
    POPULARS = -2
    RECENTS = -3
    BOOKMARK = -4
    SUGGESTION = -5
    HISTORY = -6
    SEARCH = -7
    TAG = -8
    UNCLASSIFIED = -9
    WEBVIEW = -10
    SEPARATOR = -11
    SETTINGS = -12


class LoadingState:
    NONE = 0
    LOADING = 1
    STOPPED = 2
    ERROR = 3


class Score:
    SUGGESTION = 1000
    WEBVIEW = 500


class StartPage:
    SEARCH = 0
    POPULARITY_HISTORY = 1
    POPULARITY_BOOKMARKS = 2
    BLANK = 3
    CUSTOM = 4


COLORS = {
    "aliceblue": (240, 248, 255),
    "antiquewhite": (250, 235, 215),
    "aqua": (0, 255, 255),
    "aquamarine": (127, 255, 212),
    "azure": (240, 255, 255),
    "beige": (245, 245, 220),
    "bisque": (255, 228, 196),
    "black": (0, 0, 0),
    "blanchedalmond": (255, 235, 205),
    "blue": (0, 0, 255),
    "blueviolet": (138, 43, 226),
    "brown": (165, 42, 42),
    "burlywood": (222, 184, 135),
    "cadetblue": (95, 158, 160),
    "chartreuse": (127, 255, 0),
    "chocolate": (210, 105, 30),
    "coral": (255, 127, 80),
    "cornflowerblue": (100, 149, 237),
    "cornsilk": (255, 248, 220),
    "crimson": (220, 20, 60),
    "cyan": (0, 255, 255),
    "darkblue": (0, 0, 139),
    "darkcyan": (0, 139, 139),
    "darkgoldenrod": (184, 134, 11),
    "darkgray": (169, 169, 169),
    "darkgreen": (0, 100, 0),
    "darkgrey": (169, 169, 169),
    "darkkhaki": (189, 183, 107),
    "darkmagenta": (139, 0, 139),
    "darkolivegreen": (85, 107, 47),
    "darkorange": (255, 140, 0),
    "darkorchid": (153, 50, 204),
    "darkred": (139, 0, 0),
    "darksalmon": (233, 150, 122),
    "darkseagreen": (143, 188, 143),
    "darkslateblue": (72, 61, 139),
    "darkslategray": (47, 79, 79),
    "darkslategrey": (47, 79, 79),
    "darkturquoise": (0, 206, 209),
    "darkviolet": (148, 0, 211),
    "deeppink": (255, 20, 147),
    "deepskyblue": (0, 191, 255),
    "dimgray": (105, 105, 105),
    "dimgrey": (105, 105, 105),
    "dodgerblue": (30, 144, 255),
    "firebrick": (178, 34, 34),
    "floralwhite": (255, 250, 240),
    "forestgreen": (34, 139, 34),
    "fuchsia": (255, 0, 255),
    "gainsboro": (220, 220, 220),
    "ghostwhite": (248, 248, 255),
    "gold": (255, 215, 0),
    "goldenrod": (218, 165, 32),
    "gray": (128, 128, 128),
    "green": (0, 128, 0),
    "greenyellow": (173, 255, 47),
    "grey": (128, 128, 128),
    "honeydew": (240, 255, 240),
    "hotpink": (255, 105, 180),
    "indianred": (205, 92, 92),
    "indigo": (75, 0, 130),
    "ivory": (255, 255, 240),
    "khaki": (240, 230, 140),
    "lavender": (230, 230, 250),
    "lavenderblush": (255, 240, 245),
    "lawngreen": (124, 252, 0),
    "lemonchiffon": (255, 250, 205),
    "lightblue": (173, 216, 230),
    "lightcoral": (240, 128, 128),
    "lightcyan": (224, 255, 255),
    "lightgoldenrodyellow": (250, 250, 210),
    "lightgray": (211, 211, 211),
    "lightgreen": (144, 238, 144),
    "lightgrey": (211, 211, 211),
    "lightpink": (255, 182, 193),
    "lightsalmon": (255, 160, 122),
    "lightseagreen": (32, 178, 170),
    "lightskyblue": (135, 206, 250),
    "lightslategray": (119, 136, 153),
    "lightslategrey": (119, 136, 153),
    "lightsteelblue": (176, 196, 222),
    "lightyellow": (255, 255, 224),
    "lime": (0, 255, 0),
    "limegreen": (50, 205, 50),
    "linen": (250, 240, 230),
    "magenta": (255, 0, 255),
    "maroon": (128, 0, 0),
    "mediumaquamarine": (102, 205, 170),
    "mediumblue": (0, 0, 205),
    "mediumorchid": (186, 85, 211),
    "mediumpurple": (147, 112, 216),
    "mediumseagreen": (60, 179, 113),
    "mediumslateblue": (123, 104, 238),
    "mediumspringgreen": (0, 250, 154),
    "mediumturquoise": (72, 209, 204),
    "mediumvioletred": (199, 21, 133),
    "midnightblue": (25, 25, 112),
    "mintcream": (245, 255, 250),
    "mistyrose": (255, 228, 225),
    "moccasin": (255, 228, 181),
    "navajowhite": (255, 222, 173),
    "navy": (0, 0, 128),
    "oldlace": (253, 245, 230),
    "olive": (128, 128, 0),
    "olivedrab": (107, 142, 35),
    "orange": (255, 165, 0),
    "orangered": (255, 69, 0),
    "orchid": (218, 112, 214),
    "palegoldenrod": (238, 232, 170),
    "palegreen": (152, 251, 152),
    "paleturquoise": (175, 238, 238),
    "palevioletred": (216, 112, 147),
    "papayawhip": (255, 239, 213),
    "peachpuff": (255, 218, 185),
    "peru": (205, 133, 63),
    "pink": (255, 192, 203),
    "plum": (221, 160, 221),
    "powderblue": (176, 224, 230),
    "purple": (128, 0, 128),
    "red": (255, 0, 0),
    "rosybrown": (188, 143, 143),
    "royalblue": (65, 105, 225),
    "saddlebrown": (139, 69, 19),
    "salmon": (250, 128, 114),
    "sandybrown": (244, 164, 96),
    "seagreen": (46, 139, 87),
    "seashell": (255, 245, 238),
    "sienna": (160, 82, 45),
    "silver": (192, 192, 192),
    "skyblue": (135, 206, 235),
    "slateblue": (106, 90, 205),
    "slategray": (112, 128, 144),
    "slategrey": (112, 128, 144),
    "snow": (255, 250, 250),
    "springgreen": (0, 255, 127),
    "steelblue": (70, 130, 180),
    "tan": (210, 180, 140),
    "teal": (0, 128, 128),
    "thistle": (216, 191, 216),
    "tomato": (255, 99, 71),
    "turquoise": (64, 224, 208),
    "violet": (238, 130, 238),
    "wheat": (245, 222, 179),
    "white": (255, 255, 255),
    "whitesmoke": (245, 245, 245),
    "yellow": (255, 255, 0),
    "yellowgreen": (154, 205, 50)
}



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib

from urllib.parse import urlparse

from eolie.define import Type
from eolie.popover_uri_item import Item
from eolie.popover_uri_row import Row


class UriPopoverWebviews:
    """
        Webviews handler for UriPopover
    """

    def __init__(self):
        """
            Init handler
        """
        self.__suggestion_id = None

    def add_webviews(self, value, cancellable):
        """
            Add webviews for value
            @param value as str
            @param cancellable as Gio.Cancellable
        """
        if len(value) < 2:
            return
        webviews = []
        for webview in self._window.container.webviews:
            uri = webview.uri
            if uri is None:
                continue
            parsed = urlparse(uri)
            if parsed.netloc.lower().find(value) != -1:
                webviews.append(webview)
        if webviews:
            GLib.idle_add(self.__add_webviews, webviews, cancellable)

#######################
# PRIVATE             #
#######################
    def __add_webviews(self, webviews, cancellable):
        """
            Add a row representing webview
            @param webviews as [WebView]
            @param cancellable as Gio.Cancellable
        """
        if cancellable.is_cancelled():
            return
        for webview in webviews:
            item = Item()
            item.set_property("type", Type.WEBVIEW)
            item.set_property("title", webview.title)
            item.set_property("uri", webview.uri)
            child = Row(item, self._window)
            child.show()
            self._search_box.insert(child, 0)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib

from gettext import gettext as _
from hashlib import sha256

from eolie.define import App, LoadingType
from eolie.utils import wanted_loading_type


class PagesMenu(Gio.Menu):
    """
        Menu showing closed page
    """

    def __init__(self):
        """
            Init menu
        """
        Gio.Menu.__init__(self)

        # Setup actions
        action = Gio.SimpleAction(name="openall")
        action.set_enabled(False)
        App().add_action(action)
        action.connect('activate',
                       self.__on_openall_clicked)
        # Setup menu
        self.__closed_section = Gio.Menu()
        self.append_section(_("Closed pages"), self.__closed_section)
        item = Gio.MenuItem.new(_("Open all pages"), "app.openall")
        item.set_icon(Gio.ThemedIcon.new("document-open-symbolic"))
        self.append_item(item)
        remember_session = App().settings.get_value("remember-session")
        opened_pages = App().history.get_opened_pages()
        if not remember_session and opened_pages:
            # Delayed to not slow down startup
            GLib.timeout_add(1000, self.__append_opened_pages, opened_pages)

    def activate_last_action(self):
        """
            Activate last action
        """
        if self.__closed_section.get_n_items():
            uri = self.__closed_section.get_item_attribute_value(0, "uri")
            encoded = sha256(uri.get_string().encode("utf-8")).hexdigest()
            action = App().lookup_action(encoded)
            if action is not None:
                action.activate(None)

    def add_action(self, title, uri):
        """
            Add a new action to menu
            @param title as str
            @param uri as str
        """
        # Close all item
        if not uri:
            return
        if not title:
            title = uri
        self.__clean_actions()
        encoded = sha256(uri.encode("utf-8")).hexdigest()
        action = Gio.SimpleAction(name=encoded)
        App().add_action(action)
        action.connect("activate", self.__on_action_clicked, uri)
        if len(title) > 40:
            title = title[0:40] + "…"
        item = Gio.MenuItem.new(title, "app.%s" % encoded)
        item.set_attribute_value("uri", GLib.Variant("s", uri))
        if uri == "populars://":
            item.set_icon(Gio.ThemedIcon.new("emote-love-symbolic"))
        else:
            # Try to set icon
            favicon_path = App().art.get_favicon_path(uri)
            if favicon_path is not None:
                f = Gio.File.new_for_path(favicon_path)
                icon = Gio.FileIcon.new(f)
                if icon is not None:
                    item.set_icon(icon)
            else:
                item.set_icon(Gio.ThemedIcon.new("web-browser-symbolic"))
        self.__closed_section.insert_item(0, item)
        action = App().lookup_action("openall")
        if action is not None:
            action.set_enabled(self.__closed_section.get_n_items() > 1)

    def remove_action(self, uri):
        """
            Remove action from menu
        """
        encoded = sha256(uri.encode("utf-8")).hexdigest()
        action = App().lookup_action(encoded)
        if action is not None:
            App().remove_action(encoded)
            for i in range(0, self.__closed_section.get_n_items() - 1):
                attribute = self.__closed_section.get_item_attribute_value(
                    i,
                    "uri")
                if attribute is None:
                    continue
                _uri = attribute.get_string()
                if uri == _uri:
                    self.__closed_section.remove(i)
                    break
        action = App().lookup_action("openall")
        if action is not None:
            action.set_enabled(self.__closed_section.get_n_items() < 2)

#######################
# PRIVATE             #
#######################
    def __append_opened_pages(self, opened_pages):
        """
            Append opened pages ie pages opened on last session
            @param opened_pages as [(str, str)]
        """
        for (uri, title) in opened_pages:
            self.add_action(title, uri)

    def __clean_actions(self):
        """
            Manager an history of max 20 items
        """
        if self.__closed_section.get_n_items() > 21:
            uri = self.__closed_section.get_item_attribute_value(
                0, "uri").get_string()
            encoded = sha256(uri.encode("utf-8")).hexdigest()
            action = App().lookup_action(encoded)
            if action is not None:
                App().remove_action(encoded)
            self.__closed_section.remove(0)

    def __on_openall_clicked(self, action, variant):
        """
            Add all entries
            @param Gio.SimpleAction
            @param GVariant
        """
        items = []
        for i in range(0, self.__closed_section.get_n_items()):
            uri_attr = self.__closed_section.get_item_attribute_value(i,
                                                                      "uri")
            title_attr = self.__closed_section.get_item_attribute_value(
                i,
                "label")
            if uri_attr is None or title_attr is None:
                continue
            uri = uri_attr.get_string()
            title = title_attr.get_string()
            loading_type = wanted_loading_type(i)
            items.append((uri, title, 0, False, False, None, loading_type))
            i += 1
        App().active_window.container.add_webviews(items)
        self.__closed_section.remove_all()

    def __on_action_clicked(self, action, variant, uri):
        """
            Add to playlists
            @param Gio.SimpleAction
            @param GVariant
            @param data as (str, WebKit2.WebViewSessionState)
        """
        App().active_window.container.add_webview_for_uri(
                                                  uri,
                                                  LoadingType.FOREGROUND)
        self.remove_action(uri)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, WebKit2

from urllib.parse import urlparse

from eolie.helper_passwords import PasswordsHelper
from eolie.define import App
from eolie.utils import get_baseuri, emit_signal
from eolie.logger import Logger


class WebViewHelpers:
    """
        JS helpers for webview
    """

    def __init__(self):
        """
            Init helpers
        """
        self.__readability_status = False
        self.__passwords_helper = PasswordsHelper()

    def set_forms_content(self, uuid):
        """
            Set input forms for uuid
            @parma uuid as str
        """
        self.__passwords_helper.get_by_uuid(uuid, self.__on_get_password_by)

    def check_readability(self):
        """
            Check webview readability
            @param webview as WebView
        """
        # Load Readability
        js1 = Gio.File.new_for_uri(
            "resource:///org/gnome/Eolie/Readability-readerable.js")
        js2 = Gio.File.new_for_uri(
            "resource:///org/gnome/Eolie/Readability_check.js")
        (status, content1, tags) = js1.load_contents()
        (status, content2, tags) = js2.load_contents()
        script = content1.decode("utf-8") + content2.decode("utf-8")
        self.run_javascript(script, None, self.__on_readability_status)

    @property
    def readability_status(self):
        """
            True if webview readable
            @return bool
        """
        return self.__readability_status

#######################
# PROTECTED           #
#######################
    def _on_load_changed(self, webview, event):
        """
            Run JS helpers
            @param webview as WebView
            @param event as WebKit2.LoadEvent
        """
        if event == WebKit2.LoadEvent.FINISHED:
            user_script = App().settings.get_value(
                "user-script-uri").get_string()
            f = Gio.File.new_for_uri(user_script)
            if f.query_exists():
                try:
                    (status, contents, tags) = f.load_contents(None)
                    self.run_javascript(contents.decode("utf-8"), None, None)
                except:
                    pass
            self.run_javascript_from_gresource(
                "/org/gnome/Eolie/Extensions.js", None, None)
            self.run_javascript_from_gresource(
                "/org/gnome/Eolie/javascript/HandleInput.js", None, None)
            self.__run_insecure(webview.uri)
            self.__run_set_forms(webview.uri)

#######################
# PRIVATE             #
#######################
    def __run_insecure(self, uri):
        """
            Run a script checking for password inputs while in http
            @parma uri as str
        """
        parsed = urlparse(uri)
        if parsed.scheme == "http":
            self.run_javascript_from_gresource(
                "/org/gnome/Eolie/javascript/Insecure.js", None, None)

    def __run_set_forms(self, uri):
        """
            Set input forms for current uri
            @parma uri as str
        """
        self.run_javascript_from_gresource(
                "/org/gnome/Eolie/javascript/FormMenu.js",
                None, self.__on_get_forms)

    def __on_readability_status(self, source, result):
        """
            Get readability status
            @param source as GObject.Object
            @param result as Gio.AsyncResult
        """
        try:
            data = source.run_javascript_from_gresource_finish(result)
            self.__readability_status = bool(data.get_js_value().to_string())
            emit_signal(self, "readability-status", self.__readability_status)
        except Exception as e:
            Logger.error("WebViewHelpers::__on_readability_status(): %s", e)

    def __on_get_forms(self, source, result):
        """
            Start search with selection
            @param source as GObject.Object
            @param result as Gio.AsyncResult
        """
        try:
            data = source.run_javascript_from_gresource_finish(result)
            message = data.get_js_value().to_string()
            split = message.split("\n")
            for user_form_name in split:
                self.__passwords_helper.get(get_baseuri(self.uri),
                                            user_form_name,
                                            None,
                                            self.__on_get_password_by)
        except Exception as e:
            Logger.warning("WebViewHelpers::__on_get_forms(): %s", e)

    def __on_get_password_by(self, attributes, password,
                             ignored, index, count):
        """
            Set login/password input
            @param attributes as {}
            @param password as str
            @param ignored
            @param index as int
            @param count as int
        """
        if attributes is None or count > 1:
            return
        f = Gio.File.new_for_uri(
            "resource:///org/gnome/Eolie/javascript/SetForms.js")
        (status, contents, tags) = f.load_contents(None)
        js = contents.decode("utf-8")
        js = js.replace("@INPUT_NAME@", attributes["userform"])
        js = js.replace("@INPUT_PASSWORD@", attributes["passform"])
        js = js.replace("@USERNAME@", attributes["login"])
        js = js.replace("@PASSWORD@", password)
        self.run_javascript(js, None, None)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk, GLib, Pango, Gio

from gettext import gettext as _
from urllib.parse import urlparse
from time import time

from eolie.define import App
from eolie.utils import emit_signal
from eolie.helper_task import TaskHelper
from eolie.popover_uri import UriPopover
from eolie.widget_smooth_progressbar import SmoothProgressBar
from eolie.widget_uri_entry_icons import UriEntryIcons
from eolie.helper_gestures import GesturesHelper
from eolie.helper_size_allocation import SizeAllocationHelper


class UriEntry(Gtk.Overlay, SizeAllocationHelper):
    """
        URI Entry for toolbar
    """

    def __init__(self, window):
        """
            Init toolbar
            @param window as Window
        """
        Gtk.Overlay.__init__(self)
        self.get_style_context().add_class("input")
        self.__cancellable = Gio.Cancellable.new()
        self.__window = window
        self.__task_helper = TaskHelper()
        self.__input_warning_shown = False
        self.__entry_changed_id = None
        self.__secure_content = True
        self.__size_allocation_timeout = None
        self.__current_value = ""
        self.__entry = Gtk.Entry.new()
        self.__entry.show()
        self.__entry.get_style_context().add_class("uribar")
        self.__signal_id = self.__entry.connect("changed",
                                                self.__on_entry_changed)
        self.__entry.connect("populate-popup", self.__on_entry_populate_popup)
        self.__entry.connect("icon-release", self.__on_entry_icon_release)
        self.__entry_gesture = GesturesHelper(
            self.__entry,
            primary_press_callback=self.__on_entry_press)
        self.__entry_controller = Gtk.EventControllerKey.new(self.__entry)
        self.__entry_controller.connect("focus-in", self.__on_entry_focus_in)
        self.__entry_controller.connect("focus-out", self.__on_entry_focus_out)
        self.__entry_controller.connect("key-pressed",
                                        self.__on_entry_key_pressed)

        self.__icons = UriEntryIcons(self.__entry, window)
        self.__icons.show()

        self.__placeholder = Gtk.Label.new()
        self.__placeholder.show()
        self.__placeholder.set_property("halign", Gtk.Align.START)
        self.__placeholder.set_hexpand(True)
        self.__placeholder.set_ellipsize(Pango.EllipsizeMode.END)
        self.__placeholder.get_style_context().add_class("placeholder")
        self.__placeholder.set_opacity(0.8)

        grid = Gtk.Grid()
        grid.show()
        grid.set_valign(Gtk.Align.CENTER)
        grid.add(self.__placeholder)
        grid.add(self.__icons)

        self.__progress = SmoothProgressBar()

        self.add(self.__entry)
        self.add_overlay(grid)
        self.add_overlay(self.__progress)
        self.set_overlay_pass_through(grid, True)

        # Inline completion
        self.__completion_model = Gtk.ListStore(str)
        self.__completion = Gtk.EntryCompletion.new()
        self.__completion.set_model(self.__completion_model)
        self.__completion.set_text_column(0)
        self.__completion.set_inline_completion(True)
        self.__completion.set_popup_completion(False)
        self.__entry.set_completion(self.__completion)

        self.__popover = UriPopover(window)
        self.__popover.set_relative_to(self.__entry)
        self.__popover.connect("closed", self.__on_popover_closed)

        # Some on the fly css styling
        context = self.__entry.get_style_context()
        self.__css_allocation = Gtk.CssProvider()
        self.__css_color = Gtk.CssProvider()
        context.add_provider_for_screen(Gdk.Screen.get_default(),
                                        self.__css_allocation,
                                        Gtk.STYLE_PROVIDER_PRIORITY_USER)
        context.add_provider_for_screen(Gdk.Screen.get_default(),
                                        self.__css_color,
                                        Gtk.STYLE_PROVIDER_PRIORITY_USER)
        SizeAllocationHelper.__init__(self, self.__icons)

    def set_text_entry(self, text):
        """
            Set uri in Gtk.Entry
            @param text as str
        """
        if self.__entry.get_text() == text:
            return
        if self.__signal_id is not None:
            self.__entry.disconnect(self.__signal_id)
        # Do not show this in titlebar
        parsed = urlparse(text)
        if parsed.scheme in ["populars", "about"]:
            text = ""
        # Should not be needed but set_text("") do not clear text
        self.__entry.delete_text(0, -1)
        self.__entry.set_text(text)
        self.__entry.set_position(-1)
        if text:
            self.__placeholder.set_opacity(0)
        else:
            self.__placeholder.set_opacity(0.8)
        self.__signal_id = self.__entry.connect("changed",
                                                self.__on_entry_changed)

    def set_uri(self, uri):
        """
            Set toolbar URI
            @param uri as str
        """
        self.set_tooltip_text(uri)
        self.__input_warning_shown = False
        self.__secure_content = True
        self.__update_secure_content_indicator()
        bookmark_id = App().bookmarks.get_id(uri)
        self.__icons.set_bookmarked(bookmark_id is not None)

    def set_title(self, title):
        """
            Set toolbar title
            @param title as str
        """
        if title is None:
            self.set_default_placeholder()
        else:
            self.__window.set_title(title)
            self.__placeholder.set_text(title)
            self.set_text_entry("")

    def set_insecure_content(self):
        """
            Mark uri as insecure
        """
        self.__secure_content = False
        self.__entry.set_icon_tooltip_text(
            Gtk.EntryIconPosition.PRIMARY,
            _("This page contains insecure content"))
        self.__entry.set_icon_from_icon_name(
            Gtk.EntryIconPosition.PRIMARY,
            "channel-insecure-symbolic")

    def focus(self, child="bookmarks"):
        """
            Focus wanted child entry
            @param child as str
        """
        self.get_toplevel().set_focus(self.__entry)
        uri = self.__window.container.webview.uri
        parsed = urlparse(uri)
        if parsed.scheme in ["http", "https", "file"]:
            self.set_text_entry(uri)
            self.__entry.select_region(0, -1)
        self.__popover.popup(child)

    def set_default_placeholder(self):
        """
            Show search placeholder
        """
        if self.__placeholder.get_text() == _("Search or enter address"):
            return
        self.set_text_entry("")
        self.__placeholder.set_text(_("Search or enter address"))
        self.__entry.set_icon_from_icon_name(Gtk.EntryIconPosition.PRIMARY,
                                             "system-search-symbolic")
        self.__entry.set_icon_tooltip_text(Gtk.EntryIconPosition.PRIMARY,
                                           "")

    def update_style(self):
        """
            Update color based on parent values
            Used for dark headerbar (Ubuntu like)
        """
        # Getting color for headerbar as not possible in pure CSS
        style = self.get_parent().get_style_context()
        color = style.get_color(Gtk.StateFlags.ACTIVE).to_string()
        css = ".uribar { color: %s; caret-color:%s}" % (color, color)
        self.__css_color.load_from_data(css.encode("utf-8"))

    @property
    def widget(self):
        """
            Get main entry widget
            @return Gtk.entry
        """
        return self.__entry

    @property
    def placeholder(self):
        """
            Get placeholder
            @return Gtk.Label
        """
        return self.__placeholder

    @property
    def popover(self):
        """
            Get main URI popover
            @return UriPopover
        """
        return self.__popover

    @property
    def icons(self):
        """
            Get entry icons
            @return UriEntryIcons
        """
        return self.__icons

    @property
    def progress(self):
        """
            Get progress bar
            @return Gtk.ProgressBar
        """
        return self.__progress

#######################
# PRIVATE             #
#######################
    def __focus_out(self):
        """
            Focus out widget
        """
        webview = self.__window.container.webview
        self.set_text_entry("")
        uri = webview.uri
        if uri is not None:
            bookmark_id = App().bookmarks.get_id(uri)
            self.__icons.set_bookmarked(bookmark_id is not None)

    def __update_secure_content_indicator(self):
        """
            Update PRIMARY icon, Gtk.Entry should be set
        """
        webview = self.__window.container.webview
        if time() - webview.ctime < 10:
            self.__entry.set_icon_from_icon_name(
                Gtk.EntryIconPosition.PRIMARY, "dialog-password-symbolic")
            self.__entry.set_icon_tooltip_text(
                Gtk.EntryIconPosition.PRIMARY,
                _("Save password for: %s") % webview.credentials_uri)
        else:
            uri = webview.uri
            parsed = urlparse(uri)
            self.__entry.set_icon_tooltip_text(Gtk.EntryIconPosition.PRIMARY,
                                               "")
            if (parsed.scheme == "https" and self.__secure_content) or\
                    parsed.scheme == "file":
                self.__entry.set_icon_from_icon_name(
                    Gtk.EntryIconPosition.PRIMARY,
                    "channel-secure-symbolic")
            elif parsed.scheme in ["http", "https"]:
                self.set_insecure_content()
            else:
                self.__entry.set_icon_from_icon_name(
                    Gtk.EntryIconPosition.PRIMARY,
                    "system-search-symbolic")

    def __populate_completion(self, value):
        """
            @param value as str
            @thread safe
        """
        def populate(match, value):
            if match is not None and self.__current_value == value:
                iterator = get_iterator()
                self.__completion_model.set_value(iterator,
                                                  0,
                                                  match)
                self.__completion.insert_prefix()
            else:
                # Only happen if nothing matched
                self.__completion_model.clear()

        def look_for_match(value):
            parsed = urlparse(value)
            # Look for a match in bookmarks
            match = App().bookmarks.get_match(value)
            if match is None:
                # Look for a match in history
                match = App().history.get_match(value)
            if match is not None:
                match = match.split("://")[-1].split("www.")[-1]
                # We want result to match value slashes
                slash_count = parsed.path.count("/")
                split = match.split("/")[:-slash_count - 1]
                if split:
                    match = "/".join(split)
            GLib.idle_add(populate, match, value)

        def get_iterator():
            iterator = self.__completion_model.get_iter_first()
            if iterator is None:
                iterator = self.__completion_model.insert(0)
            return iterator

        self.__task_helper.run(look_for_match, value)

    def __on_popover_closed(self, popover):
        """
            Clean titlebar
            @param popover as Gtk.popover
        """
        webview = self.__window.container.webview
        webview.grab_focus()
        self.__focus_out()
        value = self.__entry.get_text().lstrip().rstrip()
        if value:
            webview.add_text_entry(value)
        self.__entry.delete_selection()

    def __on_entry_changed(self, entry):
        """
            Delayed entry changed
            @param entry as Gtk.Entry
        """
        value = entry.get_text()
        # Block completion
        for completion in self.__completion_model:
            if completion[0] == value:
                return
        do_completion = len(value) > len(self.__current_value)
        self.__current_value = value
        self.__cancellable.cancel()
        self.__cancellable = Gio.Cancellable.new()
        if value:
            self.__placeholder.set_opacity(0)
            if not self.__popover.is_visible():
                self.__popover.popup("bookmarks")
        else:
            webview = self.__window.container.webview
            webview.clear_text_entry()
        if self.__entry_changed_id is not None:
            GLib.source_remove(self.__entry_changed_id)
        self.__entry_changed_id = GLib.timeout_add(
            100,
            self.__on_entry_changed_timeout,
            entry,
            do_completion)

    def __on_entry_changed_timeout(self, entry, do_completion):
        """
            Update popover search if needed
            @param entry as Gtk.Entry
            @param do_completion as bool
        """
        self.__entry_changed_id = None
        self.__window.container.webview.add_text_entry(self.__current_value)
        if do_completion:
            self.__populate_completion(self.__current_value)
        parsed = urlparse(self.__current_value)
        is_uri = parsed.scheme in ["about, http", "file", "https", "populars"]
        if is_uri:
            self.__popover.set_search_text(parsed.netloc + parsed.path)
        else:
            self.__popover.set_search_text(self.__current_value)
        self.__entry.set_icon_from_icon_name(Gtk.EntryIconPosition.PRIMARY,
                                             "system-search-symbolic")
        self.__entry.set_icon_tooltip_text(Gtk.EntryIconPosition.PRIMARY, "")

    def __on_entry_focus_in(self, entry, event):
        """
            Block entry on uri
            @param entry as Gtk.Entry
            @param event as Gdk.Event
        """
        if self.__popover.is_visible():
            return
        webview = self.__window.container.webview
        self.__icons.show_clear_button()
        uri = self.__window.container.webview.uri
        parsed = urlparse(uri)
        value = webview.get_current_text_entry()
        if value:
            self.set_text_entry(value)
        elif parsed.scheme in ["http", "https", "file"]:
            self.set_text_entry(uri)
            self.__placeholder.set_opacity(0)
        else:
            self.set_default_placeholder()

    def __on_entry_focus_out(self, entry, event):
        """
            Show title
            @param entry as Gtk.Entry
            @param event as Gdk.Event (do not use)
        """
        if self.__popover.is_visible():
            return
        self.__focus_out()

    def __on_entry_populate_popup(self, entry, menu):
        """
            @param entry as Gtk.Entry
            @param menu as Gtk.Menu
        """
        def on_item_activate(item, clipboard):
            self.__window.container.webview.load_uri(clipboard)
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD).wait_for_text()
        if clipboard is not None:
            item = Gtk.MenuItem.new_with_label(_("Paste and load"))
            item.connect("activate", on_item_activate, clipboard)
            item.show()
            menu.attach(item, 0, 1, 3, 4)

    def __on_entry_icon_release(self, entry, position, event):
        """
            Show popover related to icon
            @param entry as Gtk.Entry
            @param position as Gtk.EntryIconPosition
            @param event as Gdk.EventButton
        """
        if self.__entry.get_icon_name(Gtk.EntryIconPosition.PRIMARY) ==\
                "dialog-password-symbolic":
            self.__update_secure_content_indicator()
            from eolie.popover_credentials import CredentialsPopover
            credentials_popover = CredentialsPopover(self.__window)
            credentials_popover.set_relative_to(self.__entry)
            credentials_popover.set_pointing_to(
                self.__entry.get_icon_area(Gtk.EntryIconPosition.PRIMARY))
            credentials_popover.popup()
        else:
            from eolie.popover_tls import TLSPopover
            tls_popover = TLSPopover(self.__window)
            tls_popover.set_relative_to(self.__entry)
            tls_popover.set_pointing_to(
                self.__entry.get_icon_area(Gtk.EntryIconPosition.PRIMARY))
            tls_popover.popup()

    def __on_entry_press(self, x, y, event):
        """
            Show popover if hidden
            @param x as int
            @param y as int
            @param event as Gdk.EventButton
        """
        # 30 for primary icon
        if x > 30 and not self.__popover.get_visible():
            self.__on_entry_focus_in(self.__entry, event)
            self.__popover.popup("bookmarks")

    def __on_entry_key_pressed(self, controller, keyval, keycode, state):
        """
            Forward to popover history listbox if needed
            @param controller as Gtk.EventControllerKey.
            @param keyval as int
            @param keycode as int
            @param state as Gdk.ModifierType
        """
        webview = self.__window.container.webview
        uri = self.__entry.get_text().lstrip().rstrip()

        # Walk history if Ctrl + [zZ]
        if state & Gdk.ModifierType.CONTROL_MASK:
            value = None
            if keyval == Gdk.KEY_z:
                value = webview.get_prev_text_entry(uri)
            elif keyval == Gdk.KEY_Z:
                value = webview.get_next_text_entry()
            elif keyval == Gdk.KEY_Return:
                bounds = self.__entry.get_selection_bounds()
                if bounds:
                    current = self.__entry.get_text()[:bounds[0]]
                else:
                    current = self.__entry.get_text()
                value = current + ".com"
            if value is not None:
                self.set_text_entry(value)
                emit_signal(self.__entry, "changed")
            return

        # Forward event to popover, if not used, handle input
        forwarded = self.__popover.forward_event(keyval, state)
        if forwarded:
            return True
        else:
            # Close popover and save current entry
            if keyval == Gdk.KEY_Escape:
                self.set_text_entry("")
                webview.clear_text_entry()
                GLib.idle_add(self.__window.close_popovers)
                return True
            # Close popover, save current entry and load text content
            elif keyval in [Gdk.KEY_Return, Gdk.KEY_KP_Enter]:
                webview.clear_text_entry()
                GLib.idle_add(self.__window.close_popovers)
                parsed = urlparse(uri)
                # Search a missing scheme
                if uri.find(".") != -1 and\
                        uri.find(" ") == -1 and\
                        parsed.scheme not in ["http", "https"]:
                    # Add missing www.
                    if not uri.startswith("www."):
                        db_uri = App().history.get_match("www." + uri)
                        if db_uri is not None:
                            uri = "www." + uri
                    # Add missing scheme
                    db_uri = App().history.get_match(uri, True)
                    if db_uri is None:
                        uri = "http://" + uri
                    else:
                        uri = "https://" + uri
                self.__window.container.load_uri(uri)
                self.__window.container.set_expose(False)
                if self.__entry_changed_id is not None:
                    GLib.source_remove(self.__entry_changed_id)
                    self.__entry_changed_id = None
                webview.grab_focus()
                return True

    def _handle_width_allocate(self, allocation):
        """
            @param allocation as Gtk.Allocation
            @return True if allocation is valid
        """
        if SizeAllocationHelper._handle_width_allocate(self, allocation):
            style = self.__entry.get_style_context()
            if Gtk.Widget.get_default_direction() == Gtk.TextDirection.RTL:
                state = Gtk.StateFlags.NORMAL | Gtk.StateFlags.DIR_RTL
            else:
                state = Gtk.StateFlags.NORMAL | Gtk.StateFlags.DIR_LTR
            border = style.get_border(state).bottom
            padding_start = style.get_padding(state).left
            padding_end = style.get_padding(state).right
            margin_start = style.get_margin(state).left
            margin_end = style.get_margin(state).right
            margin_bottom = style.get_margin(state).bottom
            css = ".progressbar { margin-bottom: %spx;\
                   margin-left: %spx;\
                   margin-right: %spx; }" % (margin_bottom,
                                             margin_start + border,
                                             margin_end + border)
            # 5 is grid margin (see ui file)
            uribar_padding = allocation.width + 5
            css += ".uribar { padding-right: %spx; }" % (uribar_padding)
            css += ".uribar:dir(rtl)"\
                   "{ padding-left: %spx;padding-right: %spx}" %\
                   (uribar_padding, padding_end)
            # 22 is Gtk.EntryIconPosition.PRIMARY
            if Gtk.Widget.get_default_direction() == Gtk.TextDirection.RTL:
                placeholder_margin = padding_end + 22 + border
            else:
                placeholder_margin = padding_start + 22 + border
            css += ".placeholder {margin-left: %spx;}" % placeholder_margin
            css += ".placeholder:dir(rtl)"\
                   "{margin-right: %spx;\
                     margin-left: 0px;}" % placeholder_margin
            self.__css_allocation.load_from_data(css.encode("utf-8"))
            self.update_style()



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk

from math import pi


class LabelIndicator(Gtk.Label):
    """
        Label with an indicator
    """

    def __init__(self, is_count):
        """
            Init label
            @param count as bool
        """
        Gtk.Label.__init__(self)
        self.set_size_request(8, 5)
        if is_count:
            self.__margin = 0
            self.get_style_context().add_class("text-x-small")
        else:
            self.__margin = 10
        self.set_xalign(0.0)
        self.set_yalign(1)
        self.__count = 0
        self.__unshown = []

    def add(self):
        """
            Add to count
        """
        self.__count += 1
        self.__update_count()

    def remove(self):
        """
            Remove from count
        """
        self.__count -= 1
        self.__update_count()

    def mark(self, webview):
        """
            Mark view as show or unshown
            @param webview as WebView
        """
        if webview.shown and webview in self.__unshown:
            self.__unshown.remove(webview)
        elif not webview.shown and webview not in self.__unshown:
            self.__unshown.append(webview)
        self.queue_draw()

    def do_get_preferred_width(self):
        """
            Add circle width
        """
        (min, nat) = Gtk.Label.do_get_preferred_width(self)
        return (min + self.__margin, nat + self.__margin)

    def do_draw(self, cr):
        """
            Draw indicator and label
            @param cr as cairo.Context
        """
        Gtk.Label.do_draw(self, cr)
        if self.__unshown:
            w = self.get_allocated_width()
            cr.stroke()
            cr.translate(w - 4, 3)
            cr.set_line_width(1)
            Gdk.cairo_set_source_color(cr, Gdk.Color.parse("red")[1])
            cr.arc(0, 0, 2, 0, 2 * pi)
            cr.stroke_preserve()
            Gdk.cairo_set_source_color(cr, Gdk.Color.parse("red")[1])
            cr.fill()

#######################
# PRIVATE             #
#######################
    def __update_count(self):
        """
            Update view count
        """
        count = max(1, self.__count)
        if count == 1:
            self.set_text(" ")
        else:
            self.set_text(str(count))



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

from eolie.pages_manager_generic import PagesManagerGenericChild


class PagesManagerRow(Gtk.ListBoxRow, PagesManagerGenericChild):
    def __init__(self, webview, window):
        Gtk.ListBoxRow.__init__(self)
        PagesManagerGenericChild.__init__(self, webview, window)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib

from gettext import gettext as _
from urllib.parse import urlparse
import json

from eolie.helper_task import TaskHelper
from eolie.define import App, EOLIE_DATA_PATH
from eolie.logger import Logger


class Search:
    """
        Eolie search engines
    """

    def __init__(self, user_agent=None):
        """
            Init search
            @param user_agent as str
        """
        self.__user_agent = user_agent
        # Gettext does not work outside init
        self.__ENGINES = {
            'Google': [
                # Translators: Google url for your country
                _("https://www.google.com"),
                _("https://www.google.com") + '/search?q=%s&ie=utf-8&oe=utf-8',
                'https://www.google.com/complete/search?client=firefox&q=%s',
                'utf-8',
                'g'
            ],
            'DuckDuckGo': [
                'https://duckduckgo.com',
                'https://duckduckgo.com/?q=%s',
                'https://ac.duckduckgo.com/ac/?q=%s&type=list',
                'utf-8',
                'd'
            ],
            'Qwant': [
                'https://www.qwant.com',
                'https://www.qwant.com/?q=%s',
                'https://api.qwant.com/api/suggest/?q=%s&client=opensearch',
                'utf-8',
                'q'
            ],
            'Yahoo': [
                # Translators: Yahoo url for your country
                _("https://www.yahoo.com"),
                # Translators: Yahoo url for your country
                _("https://us.search.yahoo.com") + "/search?p=%s&ei=UTF-8",
                'https://ca.search.yahoo.com/sugg/ff?'
                'command=%s&output=fxjson&appid=fd',
                'utf-8',
                'y'
            ],
            'Bing': [
                'https://www.bing.com',
                'https://www.bing.com/search?q=%s',
                'https://www.bing.com/osjson.aspx?query=%s&form=OSDJAS',
                'utf-8',
                'b'
            ]
        }

        self.__engines = {}
        self.__uri = ""
        self.__search = ""
        self.__suggest = ""
        self.__encoding = ""
        self.update_default_engine()

    def save_engines(self, engines):
        """
            Save engines
            @param engines as {}
        """
        self.__engines = engines
        try:
            content = json.dumps(engines)
            f = Gio.File.new_for_path(EOLIE_DATA_PATH + "/search_engines.json")
            f.replace_contents(content.encode("utf-8"),
                               None,
                               False,
                               Gio.FileCreateFlags.REPLACE_DESTINATION,
                               None)
        except Exception as e:
            Logger.error("Search::save_engines(): %s", e)

    def update_default_engine(self):
        """
            Update default engine based on user settings
        """
        wanted = App().settings.get_value('search-engine').get_string()
        for engine in self.engines:
            if engine == wanted:
                self.__uri = self.engines[engine][0]
                self.__search = self.engines[engine][1]
                self.__suggest = self.engines[engine][2]
                self.__encoding = self.engines[engine][3]
                break

    def get_search_uri(self, words):
        """
            Return search uri for words
            @param words as str
            @return str
        """
        if len(words) > 2 and words[1] == ":":
            for engine in self.engines:
                if words.startswith("%s:" % self.engines[engine][4]):
                    return self.engines[engine][1] % words[2:]
        try:
            return self.__search % words
        except:
            return self.engines["Google"][1] % words

    def search_suggestions(self, value, cancellable, callback):
        """
            Search suggestions for value
            @param value as str
            @param cancellable as Gio.Cancellable
            @param callback as str
        """
        try:
            if not value.strip(" "):
                return
            uri = self.__suggest % GLib.uri_escape_string(value,
                                                          None,
                                                          True)
            task_helper = TaskHelper(self.__user_agent)
            task_helper.load_uri_content(uri, cancellable,
                                         callback, self.__encoding, value)
        except Exception as e:
            Logger.error("Search::search_suggestions(): %s", e)

    def install_engine(self, uri, window):
        """
            Install new search engine
            @param uri as str
            @param window as Window
        """
        self.__install_engine(uri, window)

    def is_search(self, string):
        """
            Return True is string is a search string
            @param string as str
            @return bool
        """
        # String contains space, not an uri
        search = string.find(" ") != -1 or\
            (len(string) > 2 and string[1] == ":")
        if not search:
            # String contains dot, is an uri
            search = string.find(".") == -1 and\
                string.find(":") == -1
        return search

    @property
    def engines(self):
        """
            Get engines
            return {}
        """
        if not self.__engines:
            # Load user engines
            try:
                f = Gio.File.new_for_path(EOLIE_DATA_PATH +
                                          "/search_engines.json")
                if f.query_exists():
                    (status, contents, tag) = f.load_contents(None)
                    self.__engines.update(json.loads(contents.decode("utf-8")))
            except Exception as e:
                Logger.error("Search::engines(): %s", e)
            if not self.__engines:
                self.__engines = self.__ENGINES
        return self.__engines

    @property
    def uri(self):
        """
            Search engine uri
            @return str
        """
        return self.__uri

#######################
# PRIVATE             #
#######################
    def __install_engine(self, uri, window):
        """
            Install engine from uri
            @param uri as str
            @param window as Window
        """
        task_helper = TaskHelper(self.__user_agent)
        task_helper.load_uri_content(uri, None,
                                     self.__on_engine_loaded,
                                     window)

    def __on_engine_loaded(self, uri, status, content, window):
        """
            Ask user to add engine
            @param uri as str
            @param content as bytes
            @param window as Window
        """
        SHORTNAME = "{http://a9.com/-/spec/opensearch/1.1/}ShortName"
        URL = "{http://a9.com/-/spec/opensearch/1.1/}Url"
        ENCODING = "{http://a9.com/-/spec/opensearch/1.1/}InputEncoding"
        HTML = "text/html"
        JSON = "application/x-suggestions+json"
        try:
            from eolie.popover_message import MessagePopover
            import xml.etree.ElementTree as xml
            root = xml.fromstring(content)
            name = None
            search = None
            suggest = None
            encoding = "utf-8"
            for child in root.iter():
                if child.tag == SHORTNAME:
                    name = child.text
                elif child.tag == URL:
                    if child.attrib["type"] == HTML:
                        if child.attrib["method"] == "get":
                            search = child.attrib["template"]
                    elif child.attrib["type"] == JSON:
                        suggest = child.attrib["template"]
                elif child.tag == ENCODING:
                    encoding = child.text
            if name is not None and search is not None:
                message = _("Do you want to install\n"
                            "the %s search engine?") % name
                callback = self.__on_message_popover_ok
            else:
                message = _("Unsupported search engine")
                callback = None
            popover = MessagePopover(message, window,
                                     callback,
                                     name,
                                     search,
                                     suggest or "",
                                     encoding)
            popover.set_relative_to(window.toolbar.title)
            popover.popup()
        except Exception as e:
            Logger.error("Search::__on_engine_loaded(): %s", e)

    def __on_message_popover_ok(self, name, search, suggest, encoding):
        """
            Save engine
            @param name as str
            @param uri as str
            @param suggest_uri as str
            @param encoding as str
        """
        # Save engines
        engines = App().search.engines
        parsed = urlparse(search)
        uri = "%s://%s" % (parsed.scheme, parsed.netloc)
        engines[name] = [uri,
                         search.replace("{searchTerms}", "%s"),
                         suggest.replace("{searchTerms}", "%s"),
                         encoding,
                         ""]
        App().search.save_engines(engines)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GObject, GLib, WebKit2

import json

from eolie.helper_task import TaskHelper
from eolie.utils import emit_signal
from eolie.define import EOLIE_DATA_PATH, App
from eolie.content_blocker_exceptions import ContentBlockerExceptions
from eolie.logger import Logger


class ContentBlocker(GObject.Object):
    """
        A WebKit Content Blocker
    """
    _DB_PATH = "%s/content_blocker" % EOLIE_DATA_PATH
    _JSON_PATH = "%s/content_blocker_json" % EOLIE_DATA_PATH
    __gsignals__ = {
        "set-filter": (GObject.SignalFlags.RUN_FIRST, None,
                       (GObject.TYPE_PYOBJECT,)),
        "unset-filter": (GObject.SignalFlags.RUN_FIRST, None,
                         (GObject.TYPE_PYOBJECT,))
    }

    def __init__(self, name):
        """
            Init helper
            @param blocker name as str
        """
        try:
            GObject.Object.__init__(self)
            self.__filter = None
            self.__name = name
            self.__exceptions = ContentBlockerExceptions(name)
            self._cancellable = Gio.Cancellable.new()
            self._task_helper = TaskHelper()
            self.__store = WebKit2.UserContentFilterStore.new(self._DB_PATH)
            if not GLib.file_test(self._JSON_PATH, GLib.FileTest.IS_DIR):
                GLib.mkdir_with_parents(self._JSON_PATH, 0o0750)
            if self.enabled:
                self.load()
            App().settings.connect("changed::%s" % name,
                                   self.__on_setting_changed)
        except Exception as e:
            Logger.error("ContentBlocker::__init__(): %s", e)

    def load(self):
        """
            Load from store
        """
        self.__store.load(self.__name, self._cancellable,
                          self.__on_store_load)

    def save(self, bytes):
        """
            Save to store
        """
        self.__store.save(self.__name, GLib.Bytes(bytes), self._cancellable,
                          self.__on_store_save)

    def update(self):
        """
            Update current filters with new exceptions
        """
        try:
            f = Gio.File.new_for_path(
                "%s/%s.json" % (self._JSON_PATH, self.__name))
            if f.query_exists():
                (status, content, tag) = f.load_contents(None)
                if status:
                    rules = json.loads(content.decode("utf-8"))
                    self._task_helper.run(self._save_rules, rules)
            else:
                self._task_helper.run(self._save_rules, self.DEFAULT)
        except Exception as e:
            Logger.error("ContentBlocker::update(): %s", e)

    def stop(self):
        """
            Stop update
        """
        self._cancellable.cancel()
        self._cancellable = Gio.Cancellable.new()

    @property
    def exceptions(self):
        """
            Get exceptions
            @return ContentBlockerExceptions
        """
        return self.__exceptions

    @property
    def filter(self):
        """
            Get filters
            return WebKit2.UserContentFilter
        """
        return self.__filter

    @property
    def name(self):
        """
            Get name
            return str
        """
        return self.__name

    @property
    def enabled(self):
        """
            True if enabled
            @return bool
        """
        return App().settings.get_value(self.__name)

#######################
# PROTECTED           #
#######################
    def _save_rules(self, rules):
        """
            Save rules to file
            @param uri as str
            @param rules []
        """
        try:
            new_rules = list(rules)
            if self.__exceptions is not None:
                new_rules += self.__exceptions.rules
            bytes = json.dumps(new_rules).encode("utf-8")
            self.save(bytes)
        except Exception as e:
            Logger.error("ContentBlocker::_save_rules(): %s", e)

#######################
# PRIVATE             #
#######################
    def __on_store_load(self, store, result):
        """
            Notify for new filter
            @param store as WebKit2.UserContentFilterStore
            @param result as Gio.AsyncResult
            @param encoded as str
        """
        try:
            self.__filter = store.save_finish(result)
            if self.enabled:
                emit_signal(self, "set-filter", self.__filter)
        except Exception as e:
            Logger.error("ContentBlocker::__on_store_load(): %s", e)

    def __on_store_save(self, store, result):
        """
            Notify for new filter
            @param store as WebKit2.UserContentFilterStore
            @param result as Gio.AsyncResult
        """
        try:
            self.__filter = store.load_finish(result)
            if self.enabled:
                emit_signal(self, "set-filter", self.__filter)
        except Exception as e:
            Logger.error("ContentBlocker::__on_store_save(): %s", e)

    def __on_setting_changed(self, settings, value):
        """
            Enable disable filtering
            @param settings as Gio.Settings
            @param value as GLib.Variant
        """
        if self.enabled:
            self.load()
        else:
            emit_signal(self, "unset-filter", self.__filter)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, WebKit2


class WebViewPopover(Gtk.Popover):
    """
        Show a popover with a webview stack
    """

    def __init__(self, window):
        """
            Init popover
            @param window as Window
        """
        Gtk.Popover.__init__(self)
        self.set_modal(False)
        window.register(self, False)
        self.__window = window
        builder = Gtk.Builder()
        builder.add_from_resource('/org/gnome/Eolie/PopoverWebView.ui')
        builder.connect_signals(self)
        self.__stack = builder.get_object("stack")
        self.__title = builder.get_object("title")
        self.__prev_button = builder.get_object("prev_button")
        self.__next_button = builder.get_object("next_button")
        self.__spinner = builder.get_object("spinner")
        self.__combobox = builder.get_object("combobox")
        self.__label = builder.get_object("label")
        self.add(builder.get_object("widget"))
        self.connect("closed", self.__on_closed)

    def add_webview(self, webview):
        """
            Add view to popover
            @param webview as WebView
        """
        position = len(self.__stack.get_children())
        self.__stack.add(webview)
        self.__stack.set_visible_child(webview)
        webview.connect("close", self.__on_webview_close)
        webview.connect("title-changed", self.__on_webview_title_changed)
        webview.connect("load-changed", self.__on_webview_load_changed)
        title = webview.title or webview.uri or ""
        self.__combobox.append(str(webview), title)

        size = self.__window.get_size()
        width = min(800, size[0] * 0.8)
        height = min(800, size[1] * 0.6)
        self.set_size_request(width, height)

        if position == 0:
            self.__label.set_text(title)
            self.__label.show()
            self.__combobox.set_active_id(str(webview))
            self.__combobox.hide()
        else:
            self.__label.hide()
            self.__combobox.show()

#######################
# PROTECTED           #
#######################
    def _on_combobox_changed(self, combobox):
        """
            Update visible webview
            @param combobox as Gtk.ComboBoxText
        """
        active_id = combobox.get_active_id()
        for child in self.__stack.get_children():
            if str(child) == active_id:
                self.__stack.set_visible_child(child)
                break

#######################
# PRIVATE             #
#######################
    def __get_prev_child(self):
        """
            Get previous child
            @return webview
        """
        current = self.__stack.get_visible_child()
        child = None
        for webview in self.__stack:
            if webview == current:
                break
            child = webview
        return child

    def __get_next_child(self):
        """
            Get next child
            @return webview
        """
        current = self.__stack.get_visible_child()
        at_current = False
        child = None
        for webview in self.__stack:
            if at_current:
                child = webview
                break
            if webview == current:
                at_current = True
        return child

    def __on_webview_title_changed(self, webview, title):
        """
            Update title if webview is current
            @param webview as WebView
            @param title as str
        """
        # Search previous entry
        position = 0
        for child in self.__stack.get_children():
            if webview == child:
                break
            position += 1
        self.__combobox.remove(position)
        self.__combobox.insert(position, str(webview), title)
        self.__label.set_text(title)
        visible = self.__stack.get_visible_child()
        if visible == webview:
            self.__combobox.set_active_id(str(visible))

    def __on_webview_load_changed(self, webview, event):
        """
            Update spinner
            @param webview as WebView
            @param event as WebKit2.LoadEvent
        """
        if event == WebKit2.LoadEvent.STARTED:
            self.__spinner.start()
        elif event == event == WebKit2.LoadEvent.FINISHED:
            self.__spinner.stop()

    def __on_webview_close(self, webview):
        """
            Remove view from stack, destroy it if wanted
            @param webview as WebView
            @param destroy as bool
        """
        webview.disconnect_by_func(self.__on_webview_title_changed)
        webview.disconnect_by_func(self.__on_webview_load_changed)
        for child in self.__stack.get_children():
            if child == webview:
                self.__stack.remove(child)
                break
        if self.__stack.get_children():
            pass
        else:
            self.hide()
        child.destroy()

    def __on_closed(self, popover):
        """
            Clean up popover
            Remove children
            @param popover as Gtk.Popover
        """
        for webview in self.__stack.get_children():
            webview.destroy()



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gdk, GLib, Gtk, Pango, GdkPixbuf

from math import pi
import unicodedata
import string
import cairo
from threading import current_thread
from urllib.parse import urlparse
from random import choice
from base64 import b64encode

from eolie.logger import Logger
from eolie.define import ArtSize, LoadingType


def get_char_surface(char):
    """
        Draw a char with a random color
        @param char as str
        @return cairo surface
    """
    colors = [[0.102, 0.737, 0.612],                 # Turquoise
              [0.204, 0.596, 0.859],                 # Peterriver
              [0.608, 0.349, 0.714],                 # Amethyst
              [0.204, 0.286, 0.369],                 # Wetasphalt
              [0.086, 0.627, 0.522],                 # Greensea
              [0.153, 0.682, 0.376],                 # Nephritis
              [0.161, 0.502, 0.725],                 # Belizehole
              [0.557, 0.267, 0.678],                 # Wisteria
              [0.173, 0.243, 0.314],                 # Midnightblue
              [0.827, 0.329, 0.0],                   # Pumpkin
              [0.753, 0.224, 0.169],                 # Pomegranate
              [0.498, 0.549, 0.553]                  # Asbestos
              ]
    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32,
                                 ArtSize.FAVICON,
                                 ArtSize.FAVICON)
    context = cairo.Context(surface)
    try:
        index = string.ascii_lowercase.index(char)
    except:
        index = 0
    colors_count = len(colors)
    while int(index) >= colors_count:
        index /= 2
    color = colors[int(index)]
    context.set_source_rgb(color[0], color[1], color[2])
    context.select_font_face("Sans", cairo.FONT_SLANT_NORMAL,
                             cairo.FONT_WEIGHT_BOLD)
    context.set_font_size(ArtSize.FAVICON)
    (xbearing, ybearing,
     width, height,
     xadvance, yadvance) = context.text_extents(char)
    context.move_to(ArtSize.FAVICON / 2 - (xadvance + xbearing) / 2,
                    ArtSize.FAVICON / 2 - ybearing - height / 2)
    context.show_text(char)
    context.stroke()
    return surface


def get_baseuri(uri):
    """
        Get Base URI
        @param uri as str
        @return str
    """
    parsed = urlparse(uri)
    return "%s://%s" % (parsed.scheme, parsed.netloc)


def get_safe_netloc(uri):
    """
        Get netloc (scheme if empty)
        @param uri as str
    """
    parsed = urlparse(uri)
    netloc = parsed.netloc
    if not netloc:
        netloc = "%s://" % urlparse(uri).scheme
    else:
        split = netloc.split(".")
        if len(split) > 2:
            netloc = ".".join(split[-2:])
    return netloc


def wanted_loading_type(index):
    """
        Return window type based on current index
    """
    if index == 0:
        return LoadingType.FOREGROUND
    elif index == 1:
        return LoadingType.BACKGROUND
    else:
        return LoadingType.OFFLOAD


def emit_signal(obj, signal, *args):
    """
        Emit signal
        @param obj as GObject.Object
        @param signal as str
        @thread safe
    """
    if current_thread().getName() == "MainThread":
        obj.emit(signal, *args)
    else:
        GLib.idle_add(obj.emit, signal, *args)


def get_round_surface(image, scale_factor, radius):
    """
        Get rounded surface from pixbuf
        @param image as GdkPixbuf.Pixbuf/cairo.Surface
        @return surface as cairo.Surface
        @param scale_factor as int
        @param radius as int
        @warning not thread safe!
    """
    width = image.get_width()
    height = image.get_height()
    is_pixbuf = isinstance(image, GdkPixbuf.Pixbuf)
    if is_pixbuf:
        width = width // scale_factor
        height = height // scale_factor
        radius = radius // scale_factor
    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
    ctx = cairo.Context(surface)
    degrees = pi / 180
    ctx.arc(width - radius, radius, radius, -90 * degrees, 0 * degrees)
    ctx.arc(width - radius, height - radius,
            radius, 0 * degrees, 90 * degrees)
    ctx.arc(radius, height - radius, radius, 90 * degrees, 180 * degrees)
    ctx.arc(radius, radius, radius, 180 * degrees, 270 * degrees)
    ctx.close_path()
    ctx.set_line_width(10)
    if is_pixbuf:
        image = Gdk.cairo_surface_create_from_pixbuf(image, scale_factor, None)
    ctx.set_source_surface(image, 0, 0)
    ctx.clip()
    ctx.paint()
    return surface


def is_unity():
    """
        Return True if desktop is Gnome
    """
    return GLib.getenv("XDG_CURRENT_DESKTOP") == "ubuntu:GNOME"


def on_query_tooltip(label, x, y, keyboard, tooltip):
    """
        Show label tooltip if needed
        @param label as Gtk.Label
        @param x as int
        @param y as int
        @param keyboard as bool
        @param tooltip as Gtk.Tooltip
    """
    layout = label.get_layout()
    if layout.is_ellipsized():
        tooltip.set_markup(label.get_label())
        return True


def resize_favicon(favicon):
    """
        Resize surface to match favicon size
        @param favicon as cairo.surface
        @return cairo.surface
    """
    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32,
                                 ArtSize.FAVICON,
                                 ArtSize.FAVICON)
    factor = ArtSize.FAVICON / favicon.get_width()
    context = cairo.Context(surface)
    context.scale(factor, factor)
    context.set_source_surface(favicon, 0, 0)
    context.paint()
    return surface


# TODO Use Lollypop menu builder
def update_popover_internals(widget):
    """
        Little hack to force Gtk.ModelButton to show image
        @param widget as Gtk.Widget
    """
    if isinstance(widget, Gtk.Image):
        widget.show()
    elif isinstance(widget, Gtk.Label):
        widget.set_ellipsize(Pango.EllipsizeMode.END)
        widget.set_max_width_chars(40)
        widget.set_tooltip_text(widget.get_text())
    elif hasattr(widget, "forall"):
        GLib.idle_add(widget.forall, update_popover_internals)


def get_snapshot(webview, result, callback, *args):
    """
        Set snapshot on main image
        @param webview as WebKit2.WebView
        @param result as Gio.AsyncResult
        @return cairo.Surface
    """
    ART_RATIO = 1.5  # ArtSize.START_WIDTH / ArtSize.START_HEIGHT
    try:
        snapshot = webview.get_snapshot_finish(result)
        # Set start image scale factor
        ratio = snapshot.get_width() / snapshot.get_height()
        if ratio > ART_RATIO:
            factor = ArtSize.START_HEIGHT / snapshot.get_height()
        else:
            factor = ArtSize.START_WIDTH / snapshot.get_width()
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32,
                                     ArtSize.START_WIDTH,
                                     ArtSize.START_HEIGHT)
        context = cairo.Context(surface)
        context.scale(factor, factor)
        context.set_source_surface(snapshot, factor, 0)
        context.paint()
        callback(surface, *args)
    except Exception as e:
        Logger.error("get_snapshot(): %s", e)
        callback(None, *args)


def get_random_string(size):
    """
        Get a rand string at size
        @param size as int
        return str
    """
    s = ''.join(choice(string.printable) for c in range(size))
    return b64encode(s.encode("utf-8"))[:size].decode("utf-8")


def get_current_monitor_model(window):
    """
        Return monitor model as string
        @param window as Gtk.Window
        @return str
    """
    screen = Gdk.Screen.get_default()
    display = screen.get_display()
    monitor = display.get_monitor_at_window(window.get_window())
    width_mm = monitor.get_width_mm()
    height_mm = monitor.get_height_mm()
    geometry = monitor.get_geometry()
    return "%sx%s/%sx%s" % (width_mm, height_mm,
                            geometry.width, geometry.height)


def noaccents(string):
    """
        Return string without accents
        @param string as str
        @return str
    """
    nfkd_form = unicodedata.normalize('NFKD', string)
    return u"".join([c for c in nfkd_form if not unicodedata.combining(c)])


def get_ftp_cmd():
    """
        Try to guess best ftp app
        @return app cmd as str
    """
    for app in ["filezilla", "nautilus", "thunar", "nemo", "true"]:
        cmd = GLib.find_program_in_path(app)
        if cmd is not None:
            return cmd



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib, Gtk

from eolie.logger import Logger


class NightApplication:
    """
        Auto switch to night mode base on GSD state
    """

    def __init__(self):
        """
            Initiate connexion to DBus
        """
        try:
            self.__on_night_mode_changed(self.settings)
            self.settings.connect("changed::night-mode",
                                  self.__on_night_mode_changed)
            Gio.bus_get(Gio.BusType.SESSION, None, self.__on_get_bus)
            self.__wanted_temperature = None
            settings = Gio.Settings.new(
                "org.gnome.settings-daemon.plugins.color")
            self.__wanted_temperature = settings.get_value(
                "night-light-temperature").get_uint32()
        except Exception as e:
            Logger.error("NightApplication::__init__(): %s", e)

#######################
# PRIVATE             #
#######################
    def __on_get_bus(self, source, result):
        """
            Get bus and set proxy
            @param source as GObject.Object
            @param result as Gio.AsyncResult
        """
        try:
            bus = Gio.bus_get_finish(result)
            Gio.DBusProxy.new(
                bus,
                Gio.DBusProxyFlags.GET_INVALIDATED_PROPERTIES,
                None,
                "org.gnome.SettingsDaemon.Color",
                "/org/gnome/SettingsDaemon/Color",
                "org.gnome.SettingsDaemon.Color",
                None,
                self.__on_get_proxy)
        except Exception as e:
            Logger.error("NightApplication::__on_get_bus(): %s", e)

    def __on_get_proxy(self, source, result):
        """
            Get proxy and connect signal
            @param source as GObject.Object
            @param result as Gio.AsyncResult
        """
        try:
            self.__proxy = source.new_finish(result)
            self.__proxy.connect("g-properties-changed",
                                 self.__on_property_changed)
            if self.settings.get_value("auto-night-mode"):
                value = self.__proxy.get_cached_property(
                    "Temperature").get_uint32()
                self.__on_property_changed(self.__proxy,
                                           {"Temperature": value},
                                           [])
        except Exception as e:
            Logger.error("NightApplication::__on_get_bus(): %s", e)

    def __on_property_changed(self, proxy, changed_properties,
                              invalidated_properties):
        """
            Set Eolie night mode based on current temperature
            @param proxy as Gio.DBusProxy
            @param changed_properties as {}
            @param invalidated_properties as [str]
        """
        if not self.settings.get_value("auto-night-mode"):
            return
        if "Temperature" in changed_properties.keys():
            temperature = changed_properties["Temperature"]
            night_mode = self.settings.get_value("night-mode")
            new_night_mode = temperature < self.__wanted_temperature + 100
            if night_mode != new_night_mode:
                self.settings.set_value("night-mode",
                                        GLib.Variant("b", new_night_mode))

    def __on_night_mode_changed(self, settings, *ignore):
        """
            Update GTK style
            @param settings as Settings
        """
        night_mode = settings.get_value("night-mode")
        settings = Gtk.Settings.get_default()
        settings.set_property("gtk-application-prefer-dark-theme", night_mode)
        for window in self.windows:
            GLib.idle_add(window.toolbar.title.entry.update_style)
            for webview in window.container.webviews:
                webview.night_mode()



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib

from gettext import gettext as _

from eolie.logger import Logger
from eolie.define import App, LoadingType
from eolie.widget_stack import Stack
from eolie.webview_state import WebViewState, WebViewStateStruct


class StackContainer:
    """
        Stack management for container
    """

    def __init__(self):
        """
            Init container
        """
        self.__close_timeout_id = None
        self._stack = Stack()
        self._stack.set_hexpand(True)
        self._stack.set_vexpand(True)
        self._stack.show()

    def add_webview_for_uri(self, uri, loading_type, is_ephemeral=False):
        """
            Add a webview to container with uri
            @param uri as str
            @param loading_type as Gdk.LoadingType
            @param is_ephemeral as bool
            @return WebView
        """
        state = WebViewStateStruct()
        state.uri = uri
        state.is_ephemeral = is_ephemeral
        webview = WebViewState.new_from_state(state, self._window)
        webview.show()
        self.add_webview(webview, loading_type)
        return webview

    def add_webview(self, webview, loading_type):
        """
            Add a webview to container
            @param webview as WebView
            @param loading_type as Gdk.LoadingType
        """
        webview.connect("destroy", self.__on_destroy)
        self.pages_manager.add_webview(webview)
        self.sites_manager.add_webview(webview)
        self._stack.add(webview)
        # Check for expose because we will be unable to get snapshot as
        # window is not visible
        if loading_type == LoadingType.FOREGROUND and not self.in_expose:
            self.set_visible_webview(webview)
            if webview.uri == "populars:":
                self._window.toolbar.title.entry.set_insecure_content()
            else:
                self._window.toolbar.title.entry.set_title(webview.uri)
        # Do not count container webviews as destroy may be pending on somes
        # Reason: we do not remove/destroy view to let stack animation run
        count = len(self.pages_manager.children)
        self._window.toolbar.actions.count_label.set_text(str(count))
        App().update_unity_badge()
        if loading_type == LoadingType.OFFLOAD:
            webview.set_title(webview.uri)
        elif loading_type == LoadingType.BACKGROUND or self.in_expose:
            webview.load_uri(webview.uri)

    def remove_webview(self, webview):
        """
            Remove view from container
            @param webview as WebView
        """
        self._stack.remove(webview)
        self.pages_manager.remove_webview(webview)
        self.sites_manager.remove_webview(webview)
        children = self._stack.get_children()
        if children:
            count = len(children)
            self._window.toolbar.actions.count_label.set_text(str(count))
            App().update_unity_badge()
            self.pages_manager.update_visible_child()
            self.sites_manager.update_visible_child()
        else:
            for window in App().windows:
                window.mark(False)
            self._window.close()

    def try_close_webview(self, webview):
        """
            Ask user before closing view if forms filled
            @param webview as WebView
        """
        webview.run_javascript("document.activeElement.tagName;", None,
                               self.__on_get_active_element, webview)
        self.__close_timeout_id = GLib.timeout_add(3000,
                                                   self.close_webview,
                                                   webview)

    def close_webview(self, webview):
        """
            close current webview
            @param view as View
            @param animate as bool
        """
        self.__close_timeout_id = None
        # Get children less view
        webviews = [child for child in self._stack.get_children()]
        if webview not in webviews:
            return
        webviews.remove(webview)
        webviews_count = len(webviews)
        App().history.set_page_state(webview.uri)
        self._window.close_popovers()
        # Needed to unfocus titlebar
        self._window.set_focus(None)
        was_current = webview == self._window.container.webview
        if not webview.is_ephemeral:
            App().pages_menu.add_action(webview.title, webview.uri)

        webview.destroy()
        # Don't show 0 as we are going to open a new one
        if webviews_count:
            App().update_unity_badge()
            self._window.toolbar.actions.count_label.set_text(
                str(webviews_count))

        # Nothing to do if was not current page
        if not was_current:
            return False

        next_webview = None

        # First we search for a child for current view
        if webview.children:
            next_webview = webview.children[0]

        # Current webview children not needed, clear parent
        for child in webview.children:
            child.set_parent(None)

        # Next we search for a brother for current view
        # If no brother, use parent
        parent = webview.parent
        if next_webview is None and parent is not None:
            for parent_child in parent.children:
                if webview != parent_child:
                    next_webview = parent_child
                    break
            if next_webview is None and parent in webviews:
                next_webview = parent

        # Next we search for view with higher atime
        if next_webview is None:
            atime = 0
            for webview in reversed(webviews):
                if webview.atime >= atime:
                    next_webview = webview
                    atime = webview.atime

        if next_webview is not None:
            self._window.container.set_visible_webview(next_webview)
        else:
            # We are last row, add a new one
            self.add_webview_for_uri(App().start_page, LoadingType.FOREGROUND)

    def set_visible_webview(self, webview):
        """
            Set visible webview
            @param webview as WebView
        """
        self._stack.set_visible_child(webview)
        self.pages_manager.update_visible_child()
        self.sites_manager.update_visible_child()

    @property
    def webview(self):
        """
            Get current webview
            @return webview
        """
        if self._reading_webview is None:
            return self._stack.get_visible_child()
        else:
            return self._reading_webview

#######################
# PRIVATE             #
#######################
    def __on_destroy(self, webview):
        """
            Remove webview from monitored webviews
            @param webview as WebView
        """
        self.pages_manager.remove_webview(webview)
        self.sites_manager.remove_webview(webview)

    def __on_get_active_element(self, source, result, webview):
        """
            Ask user to close view, if ok, close view
            @param source as GObject.Object
            @param result as Gio.AsyncResult
            @param webview as WebView
        """
        def on_response_id(dialog, response_id, webview, self):
            if response_id == Gtk.ResponseType.CLOSE:
                self.close_webview(webview)
            dialog.destroy()

        def on_close(widget, dialog):
            dialog.response(Gtk.ResponseType.CLOSE)

        def on_cancel(widget, dialog):
            dialog.response(Gtk.ResponseType.CANCEL)

        try:
            if self.__close_timeout_id is not None:
                GLib.source_remove(self.__close_timeout_id)
                self.__close_timeout_id = None
            data = source.run_javascript_finish(result)
            name = data.get_js_value().to_string()
            if name == "TEXTAREA":
                builder = Gtk.Builder()
                builder.add_from_resource("/org/gnome/Eolie/QuitDialog.ui")
                dialog = builder.get_object("dialog")
                label = builder.get_object("label")
                close = builder.get_object("close")
                cancel = builder.get_object("cancel")
                label.set_text(_("Do you really want to close this page?"))
                dialog.set_transient_for(self._window)
                dialog.connect("response", on_response_id, webview, self)
                close.connect("clicked", on_close, dialog)
                cancel.connect("clicked", on_cancel, dialog)
                dialog.run()
            else:
                self.close_webview(webview)
        except Exception as e:
            Logger.error("StackContainer::__on_forms_filled(): %s", e)
            self.close_webview(webview)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Gio, WebKit2, Gdk

from time import time
from gettext import gettext as _

from eolie.define import App
from eolie.logger import Logger


class DownloadRow(Gtk.ListBoxRow):
    """
        A Download row row
    """

    def __init__(self, download, finished):
        """
            Init row
            @param download as WebKit2.Download
            @param finished as bool
        """
        Gtk.ListBoxRow.__init__(self)
        self.__download = download
        self.__finished = finished
        self.__download_previous_time = None
        self.__download_bytes = 0
        self.__avg_download_rates = []
        self.__uri = self.__download.get_request().get_uri()
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/RowDownload.ui")
        builder.connect_signals(self)
        self.__preview = builder.get_object("preview")
        self.__progress = builder.get_object("progress")
        self.__label = builder.get_object("label")
        self.__sublabel = builder.get_object("sublabel")
        destination = download.get_destination()
        self.__label.set_label(destination.split("/")[-1])
        self.__button = builder.get_object("button")
        self.__button_image = builder.get_object("button_image")
        if finished:
            f = Gio.File.new_for_uri(destination)
            if f.query_exists():
                self.__on_finished(download)
            else:
                self.__button_image.set_from_icon_name("view-refresh-symbolic",
                                                       Gtk.IconSize.MENU)
                self.get_style_context().add_class("download-failed")
                self.__progress.set_opacity(0)
        else:
            progress = download.get_estimated_progress()
            self.__progress.set_fraction(progress)
        self.add(builder.get_object("row"))
        self.connect("map", self.__on_map)
        self.connect("unmap", self.__on_unmap)

    @property
    def finished(self):
        """
            True if download is finished
        """
        return self.__finished

    @property
    def download(self):
        """
            Get row download
            @return WebKit2.Download
        """
        return self.__download

#######################
# PROTECTED           #
#######################
    def _on_button_clicked(self, button):
        """
            Cancel download or restart download
            @param button as Gtk.Button
        """
        if self.__button_image.get_icon_name()[0] == "window-close-symbolic":
            self.__download.cancel()
            self.get_style_context().add_class("download-failed")
        elif self.__button_image.get_icon_name()[0] == "view-refresh-symbolic":
            webview = self.__download.get_web_view()
            App().download_manager.remove(self.__download)
            if webview is None:
                WebKit2.WebContext.get_default().download_uri(self.__uri)
            else:
                webview.download_uri(self.__uri)
            self.destroy()

#######################
# PRIVATE             #
#######################
    def __human_bytes_per_sec(self, bytes_per_sec):
        """
            Convert bytes per seconds to human visible string
            @param bytes_per_sec as float
            @return str
        """
        prefix = _("Download speed:")
        if bytes_per_sec < 1024:
            suffix = _("bytes/s")
            string = "%s" % round(bytes_per_sec, 2)
        elif bytes_per_sec / 1024 < 1024:
            suffix = _("KiB/s")
            string = "%s" % round(bytes_per_sec / 1024, 2)
        else:
            suffix = _("MiB/s")
            string = "%s" % round(bytes_per_sec / 1024 / 1024, 2)
        return "%s\n%s %s %s" % (self.__label.get_text(),
                                 prefix,
                                 string,
                                 suffix)

    def __human_seconds(self, seconds):
        """
            Convert time in seconds to human visible string
            @param seconds as int
            @return str
        """
        seconds_str = ""
        minutes_str = ""
        hours_str = ""
        hours = -1
        minutes = -1
        seconds_wanted = True
        # Make seconds string
        if seconds < 60:
            if seconds < 2:
                seconds_str = _("%s second") % seconds
            else:
                seconds_str = _("%s seconds") % seconds
        else:
            _seconds = seconds % 60
            if _seconds < 2:
                seconds_str = _("%s second") % _seconds
            else:
                seconds_str = _("%s seconds") % _seconds

        # Make minutes string
        if seconds > 59:
            minutes = int(seconds / 60)
            if minutes < 2:
                minutes_str = _("%s minute") % minutes
            elif minutes < 60:
                seconds_wanted = False
                minutes_str = _("%s minutes") % minutes
            else:
                seconds_wanted = False
                _minutes = minutes % 60
                if _minutes < 2:
                    minutes_str = _("%s minute") % _minutes
                else:
                    minutes_str = _("%s minutes") % _minutes

        # Make hour string
        if minutes > 59:
            hours = int(minutes / 60)
            if hours < 2:
                hours_str = _("%s hours") % hours
            else:
                hours_str = _("%s hour") % hours
        string = _("Remaining time:") + " "
        if hours_str:
            string += hours_str + ", "
        if minutes_str:
            string += minutes_str
        if seconds_wanted and minutes_str:
            string += ", "
        if seconds_wanted and seconds_str:
            string += seconds_str
        return string

    def __on_map(self, widget):
        """
            Connect signals
            @param widget as Gtk.Widget
        """
        self.__download_previous_time = None
        self.__download.connect("finished", self.__on_finished)
        self.__download.connect("received-data", self.__on_received_data)
        self.__download.connect("failed", self.__on_failed)
        response = self.__download.get_response()
        if response is None:
            return
        destination = self.__download.get_destination()
        try:
            f = GLib.filename_from_uri(destination)[0]
            (mime, uncertain) = Gio.content_type_guess(f, None)
            if uncertain:
                self.__preview.set_from_icon_name("text-x-generic",
                                                  Gtk.IconSize.DIALOG)
            else:
                icon = Gio.content_type_get_icon(mime)
                self.__preview.set_from_gicon(icon, Gtk.IconSize.DIALOG)
        except:
            self.__preview.set_from_icon_name("text-x-generic",
                                              Gtk.IconSize.DIALOG)

    def __on_unmap(self, widget):
        """
            Disconnect signals
            @param widget as Gtk.Widget
        """
        self.__download.disconnect_by_func(self.__on_finished)
        self.__download.disconnect_by_func(self.__on_received_data)
        self.__download.disconnect_by_func(self.__on_failed)

    def __on_received_data(self, download, length):
        """
            @param download as WebKit2.Download
            @param length as int
        """
        try:
            response = download.get_response()
            incoming = response.get_content_length() -\
                download.get_received_data_length()
            new_time = time()
            self.__download_bytes += length
            if self.__download_previous_time is not None:
                delta = new_time - self.__download_previous_time
                # Update every 1 seconds
                if delta > 1:
                    bytes_per_second = 1 * self.__download_bytes /\
                        delta
                    self.__avg_download_rates.append(bytes_per_second)
                    self.set_tooltip_text(self.__human_bytes_per_sec(
                        bytes_per_second))
                    # Calculate average for last 10 rates
                    avg = 0
                    for rate in self.__avg_download_rates:
                        avg += rate
                    if len(self.__avg_download_rates) > 10:
                        self.__avg_download_rates.pop(0)
                    avg /= len(self.__avg_download_rates)
                    if incoming > 0:
                        seconds = incoming / avg
                        self.__sublabel.set_label(
                            self.__human_seconds(int(seconds)))
                    self.__download_bytes = 0
                    self.__download_previous_time = new_time
            else:
                self.__download_previous_time = new_time
            self.__progress.set_fraction(download.get_estimated_progress())
        except Exception as e:
            Logger.error("DownloadRow::__on_received_data(): %s", e)

    def __on_finished(self, download):
        """
            @param download as WebKit2.Download
        """
        self.__sublabel.set_label("")
        self.set_tooltip_text(self.__label.get_text())
        self.__finished = True
        self.__progress.set_opacity(0)
        parent = self.get_parent()
        if parent is not None:
            parent.invalidate_sort()
        if self.__button_image.get_icon_name()[0] == "view-refresh-symbolic":
            return True
        else:
            self.__button.hide()

    def __on_failed(self, download, error):
        """
            @param download as WebKit2.Download
            @param error as GLib.Error
        """
        self.__button_image.set_from_icon_name("view-refresh-symbolic",
                                               Gtk.IconSize.MENU)
        self.get_style_context().add_class("download-failed")


class DownloadsPopover(Gtk.Popover):
    """
        Show current downloads
    """

    def __init__(self, window):
        """
            Init popover
            @param window as Window
        """
        Gtk.Popover.__init__(self)
        self.__window = window
        self.set_modal(False)
        window.register(self)
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/PopoverDownloads.ui")
        builder.connect_signals(self)
        self.__listbox = builder.get_object("downloads_box")
        self.__listbox.connect("row-activated", self.__on_row_activated)
        self.__listbox.set_placeholder(builder.get_object("placeholder"))
        self.__scrolled = builder.get_object("scrolled")
        self.__clear_button = builder.get_object("clear_button")
        self.add(builder.get_object("widget"))
        self.connect("map", self.__on_map)
        self.connect("unmap", self.__on_unmap)
        self.__populate()

#######################
# PROTECTED           #
#######################
    def _on_open_clicked(self, button):
        """
            Open download folder
            @param button as Gtk.button
        """
        directory_uri = App().settings.get_value("download-uri").get_string()
        if not directory_uri:
            directory = GLib.get_user_special_dir(
                GLib.UserDirectory.DIRECTORY_DOWNLOAD)
            directory_uri = GLib.filename_to_uri(directory, None)
        Gtk.show_uri_on_window(self.__window,
                               directory_uri,
                               Gdk.CURRENT_TIME)
        self.hide()

    def _on_clear_clicked(self, button):
        """
            Clear finished downloads
            @param button as Gtk.button
        """
        self.popdown()
        if len(App().download_manager.get()) == 0:
            self.__window.toolbar.end.download_button.hide()
        for download in App().download_manager.get_finished():
            App().download_manager.remove(download)

#######################
# PRIVATE             #
#######################
    def __populate(self):
        """
            Populate listbox
        """
        clear = False
        for download in reversed(App().download_manager.get()):
            if download.get_destination() is None:
                continue
            child = DownloadRow(download, False)
            child.connect("size-allocate",
                          lambda x, y: self.__calculate_height())
            child.connect("destroy",
                          lambda x: self.__calculate_height())
            child.show()
            self.__listbox.add(child)
        for download in reversed(App().download_manager.get_finished()):
            if download.get_destination() is None:
                continue
            child = DownloadRow(download, True)
            child.connect("size-allocate",
                          lambda x, y: self.__calculate_height())
            child.connect("destroy",
                          lambda x: self.__calculate_height())
            clear = True
            child.show()
            self.__listbox.add(child)
        self.__clear_button.set_sensitive(clear)

    def __calculate_height(self):
        """
            Update popover height request
        """
        height = 0
        for child in self.__listbox.get_children():
            if child.is_visible():
                height += child.get_allocated_height()
        size = self.get_ancestor(Gtk.Window).get_size()
        if height > size[1] * 0.6:
            height = size[1] * 0.6
        self.__scrolled.set_size_request(400, height)

    def __on_row_activated(self, listbox, row):
        """
            Launch row if download finished
            @param listbox as Gtk.ListBox
            @param row as Row
        """
        try:
            if row.finished:
                Gtk.show_uri_on_window(self.__window,
                                       row.download.get_destination(),
                                       Gdk.CURRENT_TIME)
                self.hide()
        except Exception as e:
            Logger.error("DownloadsPopover::__on_row_activated(): %s", e)

    def __on_map(self, widget):
        """
            Setup widget
            @param widget as Gtk.Widget
        """
        App().download_manager.connect("download-start",
                                       self.__on_download_start)
        App().download_manager.connect("download-finish",
                                       self.__on_download_finish)
        self.set_size_request(400, -1)

    def __on_unmap(self, widget):
        """
            Clean up widget
            @param widget as Gtk.Widget
        """
        for child in self.__listbox.get_children():
            child.destroy()
        App().download_manager.disconnect_by_func(self.__on_download_start)
        App().download_manager.disconnect_by_func(self.__on_download_finish)

    def __on_download_start(self, download_manager, download_name):
        """
            Add download
            @param download manager as Download Manager
            @param download_name as str
        """
        for download in App().download_manager.get():
            if download.get_destination() is None:
                continue
            if str(download) == download_name:
                child = DownloadRow(download, False)
                child.connect("size-allocate",
                              lambda x, y: self.__calculate_height())
                child.connect("destroy",
                              lambda x: self.__calculate_height())
                child.show()
                self.__listbox.add(child)
                break

    def __on_download_finish(self, download_manager):
        """
            Update clear button
            @param download manager as Download Manager
        """
        self.__clear_button.set_sensitive(True)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import json

from eolie.content_blocker import ContentBlocker
from eolie.logger import Logger


class MediasContentBlocker(ContentBlocker):
    """
        A WebKit Content Blocker for medias
    """

    DEFAULT = [
        {
            "trigger": {
                "url-filter": ".*",
                "resource-type": ["media"]
            },
            "action": {
                "type": "block"
            }
        }
    ]

    def __init__(self):
        """
            Init adblock helper
        """
        try:
            ContentBlocker.__init__(self, "block-medias")
            rules = self.DEFAULT + self.exceptions.rules
            bytes = json.dumps(rules).encode("utf-8")
            self.save(bytes)
        except Exception as e:
            Logger.error("PopupsContentBlocker::__init__(): %s", e)
